/**
 * @file Application.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-13
 */

#pragma once

#include <string_view>
#include "Sdl++/Metrics.h"
#include "Sdl++/SDL++.h"
#include "Screen.h"
#include "Sdl++/SdlWindow.h"
#include "Sdl++/Context.h"
#include "Sdl++/XDGFilePaths.h"

namespace rose {

    /**
     * @class Application
     * @brief An abstraction of the application interface with the display screen.
     */
    class Application {
    public:
        enum ApplicationVisualState {
            MINIMIZED, RESTORED, MAXIMIZED, DESKTOP,
        };

        using SizeCallbackType = Size<int>(*)(const Size<int> &); //std::function<Size<int>(const Size<int>&)>;

    protected:
        static constexpr std::string_view TEST_OPTION = "--TEST";
        static constexpr std::string_view THEME_OPTION = "--theme";
        static constexpr std::array<XDGFilePaths::XDG_Name, 3> confSearch{
                XDGFilePaths::XDG_CONFIG_DIRS, XDGFilePaths::XDG_CONFIG_HOME, XDGFilePaths::XDG_DATA_DIRS};
        std::unique_ptr<InputParser> mInputParser{};
        std::shared_ptr<Screen> mScreen{};
        std::vector<std::shared_ptr<Popup>> mPopupList{};
        gm::Context mContext{};
        gm::SdlWindow sdlWindow{};
        std::atomic_bool mRunEventLoop{true};
        SDL_version sdlVersion{};
        int mNumberDisplays{};
        std::function<void(SDL_Event)> eventCallback{};
        std::size_t mFrame{};
        gm::Texture mBaseTexture{};

        Size<int> sdlWindowSize{};
        Position<int> sdlWindowPosition{UndefinedPosition};
        Position<int> mousePosition{};

        std::vector<Rectangle<int>> mDisplayBounds{};

        ApplicationVisualState applicationVisualState{RESTORED};

        KeyShortcutList mKeyShortcutList{};

        SizeCallbackType mSizeRequestCallback{};

        void draw();

        Application() = default;

        void completeInitialize();

    public:

        static Application& get() {
            static Application application{};
            return application;
        }

        template<class...Args>
        void initialize(int argc, char **argv, Args...args) {
            mInputParser = std::make_unique<InputParser>(argc, argv);
            (setParameter(std::forward<Args>(args)), ...);
            completeInitialize();
        }

        template<class Arg>
        void setParameter(Arg arg) {
            throw std::runtime_error("Parameter not supported by Application.");
        }

        ~Application();

        void eventLoop();

        void basicEventCallback(SDL_Event e);

        [[nodiscard]] Size<int> getWindowSize() const {
            int w, h;
            SDL_GetWindowSize(sdlWindow.get(), &w, &h);
            return Size<int>{w, h};
        }

        std::shared_ptr<Screen> screen() { return mScreen; }

        void addPopup(const std::shared_ptr<Popup>& popup) {
            mPopupList.push_back(popup);
        }

        void removePopup(const std::shared_ptr<Popup>& popup) {
            mPopupList.erase(std::remove_if(mPopupList.begin(), mPopupList.end(), [&popup](const std::shared_ptr<Popup>& p){
                return p == popup;
            }), mPopupList.end());
        }

        void minimize();

        void restore();

        void desktop();
    };

    template<>
    inline void Application::setParameter<Size<int>>(Size<int> arg) {
        sdlWindowSize = arg;
    }

    template<>
    inline void Application::setParameter<Position<int>>(Position<int> arg) {
        sdlWindowPosition = arg;
    }

    template<>
    inline void Application::setParameter<Application::SizeCallbackType>(SizeCallbackType arg) {
        mSizeRequestCallback = arg;
    }
}

