/**
 * @file Builder.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-25
 */

#pragma once

#include "Application.h"
#include "Container.h"

namespace rose {

    /**
     * @class Builder
     * @brief A helper class and function to build widgets using simple semantics.
     */
    template <class WidgetClass>
    class WidgetBuilder {
    protected:
        std::shared_ptr<WidgetClass> widget{};

    public:

        WidgetBuilder() {
            static_assert(std::is_base_of_v<Widget,WidgetClass>, "Builder WidgetClass must derive from Widget.");
            widget = std::make_shared<WidgetClass>();
            widget->buildComplete();
        }

        template<typename...Args>
        explicit WidgetBuilder(Args...args) {
            widget = std::make_shared<WidgetClass>();
            ((widget->template setParameter(args)),...);
            widget->buildComplete();
        }

        ~WidgetBuilder() = default;

        std::shared_ptr<WidgetClass> getWidget() { return widget; }

        std::shared_ptr<WidgetClass> create() {
            return widget;
        }

        template<class ContainerClass>
        void addToBuilder(WidgetBuilder<ContainerClass>& container) {
            auto c = std::dynamic_pointer_cast<Container>(container.getWidget());
            if (c)
                c->addWidget(widget);
            else
                throw std::runtime_error("Must add a container to a builder.");
        }

        void addToContainer(const std::shared_ptr<Container>& container) {
            container->addWidget(widget);
            widget->insertedInTree();
        }

        void popup() {
            static_assert(std::is_base_of_v<Popup,WidgetClass>, "Only Popup derived objects can popup.");
            Application::get().addPopup(widget);
        }

        template<class Parameter>
        void set(Parameter parameter) {
            widget->setParameter(parameter);
        }
    };

    template<class WidgetClass>
    WidgetBuilder<WidgetClass> Builder() {
        WidgetBuilder<WidgetClass> widgetBuilder{};
        return widgetBuilder;
    }

    template<class WidgetClass, typename...Args>
    WidgetBuilder<WidgetClass> Builder(Args...args) {
        WidgetBuilder<WidgetClass> widgetBuilder{args...};
        return widgetBuilder;
    }

    template<class PopupClass>
    WidgetBuilder<PopupClass> PopupBuilder() {
        WidgetBuilder<PopupClass> popupBuilder{};
        return popupBuilder;
    }

    template<class PopupClass, typename...Args>
    WidgetBuilder<PopupClass> PopupBuilder(Args...args) {
        WidgetBuilder<PopupClass> popupBuilder{args...};
        return popupBuilder;
    }
}

/**
 * @brief The parenting operator, add the built widget to a Container.
 * @details The '+' operator is chosen for its precedence and that it is not an unary operator.
 * @tparam WidgetClass The widget class being built.
 * @param builder The builder object.
 * @param container The container object.
 * @return builder
 */
template<class WidgetClass>
inline rose::WidgetBuilder<WidgetClass> operator + (rose::WidgetBuilder<WidgetClass> builder, const std::shared_ptr<rose::Container> &container) {
    builder.addToContainer(container);
    return std::move(builder);
}

/**
 * @brief The parenting operator, add the built widget to a Container held in a WidgetBuilder.
 * @details The '+' operator is chosen for its precedence and that it is not an unary operator.
 * @tparam WidgetClass The widget class being built.
 * @tparam ContainerClass The container class built by another builder.
 * @param builder The builder holding the widget being built.
 * @param container The builder holding the container.
 * @return the builder.
 */
template<class WidgetClass, class ContainerClass>
inline rose::WidgetBuilder<WidgetClass> operator + (rose::WidgetBuilder<WidgetClass> builder, rose::WidgetBuilder<ContainerClass> &container) {
    static_assert(std::is_base_of_v<rose::Container, ContainerClass>, "ContainerClass must be derived from rose::Container.");
    builder.template addToBuilder(container);
    return std::move(builder);
}

/**
 * @brief Assign the contents of a widget builder to a widget pointer.
 * @tparam WidgetClass The class of the widget built.
 * @param widgetPtr The pointer the widget is being assigned to
 * @param builder The builder.
 * @return The builder.
 */
template<class WidgetClass>
inline rose::WidgetBuilder<WidgetClass> operator << (std::shared_ptr<WidgetClass> &widgetPtr,rose::WidgetBuilder<WidgetClass> builder) {
    widgetPtr = builder.getWidget();
    return std::move(builder);
}
