/**
 * @file Button.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-04
 */

#pragma once

#include <functional>
#include <utility>
#include "Widget.h"
#include "Sdl++/Signals.h"
#include "Text.h"

namespace rose {

    using PushButtonProtocol = Protocol<bool>;

    class Button;
    using PushButtonCallback = std::function<void(bool)>;

    /**
     * @class ButtonBehavior
     * @brief A base class for button behavior specifiers.
     * @details A behavior specifier is an add on to a button which determines its behavior. The
     * default is push button behavior.
     */
    class ButtonBehavior {
        friend class Button;

    protected:
        std::shared_ptr<Button> mButton{};

        bool& sensitive();
        bool& active();
        bool& abortOnLeave();
        PushButtonCallback callback();
        PushButtonProtocol::signal_type signal();

    public:
        ButtonBehavior() = default;

        explicit ButtonBehavior(std::shared_ptr<Widget> ptr) {
            mButton = ptr->as<Button>();
        }

        virtual ~ButtonBehavior() = default;

        virtual bool handleEvent(const Event& event);
    };

    /**
     * @class ToggleButtonBehavior
     * @brief Provide toggle behavior for a button.
     */
    class ToggleButtonBehavior : public ButtonBehavior {
    public:
        ToggleButtonBehavior() = default;

        ~ToggleButtonBehavior() override = default;

        explicit ToggleButtonBehavior(std::shared_ptr<Widget> ptr) : ButtonBehavior(ptr) {}

        bool handleEvent(const Event& event) override;
    };

    /**
     * @class Button
     * @brief
     */
    class Button : public Widget {
    public:
        friend class ButtonBehavior;

        PushButtonCallback mCallback{};

        PushButtonProtocol::signal_type pushButtonSignal{};

        std::shared_ptr<ButtonBehavior> mButtonBehavior{};

    protected:
        static constexpr std::string_view PushButtonClass{"Button"};
        FontPointSize mFontPointSize{};
        FontSelection mFontSelection{FontSelection::FIXED_BOLD};
        FontSelection mIconFontSelection{FontSelection::COUNT};
        uint32_t mIconCodePoint{};
        std::string mText{};

        LineTypeSetter lineTypeSetter{};

    private:
        void getResources();

    public:
        Button() : Widget(property::Class, PushButtonClass), mButtonBehavior() {
            getResources();
        }

        Button(const property_key &key, const property_value &value) : Widget(key, value), mButtonBehavior() {
            getResources();
        }

        ~Button() override = default;

        template<class Parameter>
        void setParameter(Parameter parameter) {
            Widget::setParameter(parameter);
        }

        template<class Parameter>
        requires string_value<Parameter>
        void setParameter(Parameter parameter) {
            mText = parameter;
        }

        template<class Parameter>
        requires pointer_base_of<ButtonBehavior,Parameter>
        void setParameter(Parameter parameter) {
            mButtonBehavior = std::move(parameter);
        }

        void buildComplete() override;

        void
        draw(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize) override;

        void
        layout(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize) override;

        bool handleEvent(const Event& event) override;
    };

    template<>
    inline void Button::setParameter<FontSize>(FontSize parameter) {
        mFontPointSize.pt = mTheme.fontSize[asSize(parameter)];
    }

    template<>
    inline void Button::setParameter<FontPointSize>(FontPointSize parameter) {
        mFontPointSize.pt = parameter.pt;
    }

    template<>
    inline void Button::setParameter<FontSelection>(FontSelection parameter) {
        if (parameter == FontSelection::ENTYPO)
            mIconFontSelection = parameter;
        else
            mFontSelection = parameter;
    }

    template<>
    inline void Button::setParameter<uint32_t>(uint32_t parameter) {
        mIconCodePoint = parameter;
    }
}

