/**
 * @file ClipRectangleGuard.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-20
 */

#pragma once

#include <SDL.h>
#include "Metrics.h"

namespace rose::gm {

    class Context;

    /**
     * @class ClipRectangleGuard
     * @brief Store the current clip rectangle replacing it with a new clip rectangle. When the object is
     * destroyed (by going out of scope) the old clip rectangle is set.
     */
    class [[maybe_unused]] ClipRectangleGuard {
    protected:
        Context &mContext;    ///< The renderer to which the clip rectangles are set.
        SDL_Rect mOldClip{};    ///< The old clip rectangle
        bool mStatus;

    public:
        ClipRectangleGuard() = delete;  ///< Deleted default constructor
        ClipRectangleGuard(const ClipRectangleGuard &) = delete;     ///< Deleted copy constructor

        explicit operator bool() const { return mStatus; }

        /**
         * @brief Set the old clip rectangle back on the renderer when destroyed.
         */
        ~ClipRectangleGuard();

        /**
         * @brief Speculative constructor.
         * @details This constructor saves the current clip rectangle to be restored later, but does not
         * set a new clip rectangle. A new clip rectangle may be set using the assignment operators.
         * @param context The renderer to guard the clip rectangle of.
         */
        [[maybe_unused]] explicit ClipRectangleGuard(Context &context);

        /**
         * @brief Constructor. Store the current clip rectangle and set the new one.
         * @param context The renderer to set the clip rectangles on.
         * @param clip The new clip rectangle.
         */
        ClipRectangleGuard(Context &context, const SDL_Rect &clip);

        /**
         * @brief Constructor. Store the current clip rectangle and set the new one.
         * @param renderer The renderer to set the clip rectangles on.
         * @param x X co-ordinate of the new clip rectangle.
         * @param y Y co-ordinate of the new clip rectangle.
         * @param w Width of the new clip rectangle.
         * @param h Height of the new clip rectangle.
         */
        [[maybe_unused]] ClipRectangleGuard(Context &context, int x, int y, int w, int h)
                : ClipRectangleGuard(context, SDL_Rect{x, y, w, h}) {}

        /**
         * @brief Conditional constructor.
         * @details This constructor saves the current clip rectangle to be restored later. If clip
         * is valid it is converted to an SDL_Rect and set as the new clip rectangle
         * @param renderer The renderer to set the clip rectangles on.
         * @param clip A, possibly invalid, RectangleInt.
         */
        [[maybe_unused]] ClipRectangleGuard(Context &context, const Rectangle<int> &clip);

        /**
         * @brief Assign a new clip rectangle through the ClipRectangleGuard.
         * @param clip The new clip rectangle.
         * @return The ClipRectangleGuard.
         */
        ClipRectangleGuard &operator=(SDL_Rect &clip);

        /**
         * @brief Assign a new clip rectangle through the ClipRectangleGuard.
         * @details If clip is invalid no changes are made.
         * @param clip The new clip rectangle.
         * @return The ClipRectangleGuard.
         */
        ClipRectangleGuard &operator=(Rectangle<int> &clip);

        /**
         * @brief Sets the clip rectangle to the intersection of the current clip rectangle and clip. If there is
         * no current clip rectangle, or there is not intersection, the clip rectangle is set to clip.
         * @param clip The argument rectangle.
         * @return false on error.
         */
        [[maybe_unused]] bool intersection(Rectangle<int> &clip);
    };

}

