/**
 * @file Container.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-02
 */

#pragma once

#include "Event.h"
#include "Widget.h"
#include "Layout.h"

namespace rose {

    /**
     * @class Container
     * @brief A base class for a functional UI object that is the trunk for other UI objects.
     */
    class Container : public Widget {
    public:
        using container_type = std::vector<std::shared_ptr<Widget>>;
        using size_type [[maybe_unused]] = container_type::size_type;
        static constexpr std::string_view ContainerClass = "Container";

    protected:
        container_type mBranches{};
        std::shared_ptr<Layout> mLayout{};

    private:
        void getResources();

    public:

        Container() : Widget(property::Class, ContainerClass) {
            getResources();
        }

        Container(const property_key key, const property_value &value) : Widget(key,value) {
            getResources();
        }

        ~Container() override = default;

        void clearPerFrameProcess() override;

        void perFrameProcess() override;

        template<class Parameter>
        void setParameter(Parameter parameter) {
            Widget::setParameter(parameter);
        }

        template<class LayoutType>
        inline void setParameter(const std::shared_ptr<LayoutType> parameter) {
            if constexpr(std::is_base_of_v<Layout, LayoutType>) {
                setLayout(parameter);
            } else {
                Widget::setParameter(parameter);
            }
        }

        std::optional<container_type> findEventWidget(const Event::EventType eventType, const Position<int> &eventPosition) {
            if (containsRelative(eventPosition)) {
                for (auto &widget: mBranches) {
                    if (widget->containsRelative(eventPosition - position)) {
                        if (auto container = std::dynamic_pointer_cast<Container>(widget); container) {
                            auto eventChain = container->findEventWidget(eventType, eventPosition - position);
                            if (!eventChain.has_value()) {
                                eventChain = container_type();
                                eventChain->push_back(widget);
                                eventChain->push_back(shared_from_this());
                                return eventChain;
                            } else {
                                eventChain->push_back(shared_from_this());
                                return eventChain;
                            }
                        } else {
                            container_type eventChain{};
                            eventChain.push_back(widget);
                            eventChain.push_back(shared_from_this());
                            return eventChain;
                        }
                    }
                }
            }
            return std::nullopt;
        }

        void setLayout(const std::shared_ptr<Layout> &layout) {
            mLayout = layout;
        }

        auto addWidget(const std::shared_ptr<Widget>& widget) {
            mBranches.push_back(widget);
            widget->setTrunk(shared_from_this());
            return shared_from_this();
        }

        bool removeWidget(const std::shared_ptr<Widget>& widget) {
            if (auto itr = std::find(mBranches.begin(), mBranches.end(), widget); itr != mBranches.end()) {
                mBranches.erase(itr);
                widget->unsetTrunk();
                return true;
            }
            return false;
        }

        void clear() {
            for (auto &sub : mBranches) {
                if (auto container = std::dynamic_pointer_cast<Container>(sub); container) {
                    container->clear();
                }
            }
            mBranches.clear();
        }

        bool empty() const { return mBranches.empty(); }

        auto view() { return mBranches; }

        auto operator[](std::size_t n) { return mBranches.operator[](n); }

        void draw(gm::Context &context, const Position<int>& trunkPosition, const Size<int>& trunkSize) override;

        void layout(gm::Context &context, const Position<int> &trunkPosition,
                    const Size<int> &trunkSize) override;

        bool handleKeyShortcut(const KeyShortcut &keyShortcut) override {
            if (Widget::handleKeyShortcut(keyShortcut))
                return true;

            for (auto &widget: mBranches) {
                if (widget->handleKeyShortcut(keyShortcut))
                    return true;
            }

            return false;
        }
    };

    using ContainerPtr = std::shared_ptr<Container>;
}
