/**
 * @file Context.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-20
 */

#pragma once

#include <memory>
#include <SDL.h>
#include "SdlWindow.h"
#include "Texture.h"
#include "Color.h"
#include "DrawColorGuard.h"

namespace rose::gm {
    /**
     * @struct RenderFlip
     * @brief A Widget manipulator to indicate if and how rendering a texture should be flipped.
     * @details The structure supports flipping on the Horizontal, Vertical or both axis.
     */
    struct RenderFlip {
        SDL_RendererFlip mFlip;

        /// Default constructor -- No flipping
        constexpr RenderFlip() noexcept: mFlip(SDL_FLIP_NONE) {}

        /// Constructor -- user specified flipping.
        [[maybe_unused]] constexpr explicit RenderFlip(SDL_RendererFlip flip) noexcept: mFlip(flip) {}
    };

    /**
     * @classs Context
     * @brief An abstraction of graphics rendering context.
     */
    class Context {
        friend class RenderTargetGuard;

    protected:
        /**
         * @brief A functor to destroy an SDL_Renderer
         */
        class RendererDestroy {
        public:
            /**
             * @brief Call the SDL API to destroy the renderer.
             * @param sdlRenderer The SDL_Renderer
             */
            void operator()(SDL_Renderer *sdlRenderer) {
                if (sdlRenderer) {
                    SDL_DestroyRenderer(sdlRenderer);
                }
            }
        };

        using RendererPtr = std::unique_ptr<SDL_Renderer, RendererDestroy>; ///< An SDL_Renderer unique pointer
        RendererPtr mRenderer{};    ///< The Renderer.

        SDL_Texture *mCurrentRenderTarget{nullptr};

    public:

        Context() = default;

        Context(const Context &context) = delete;

        Context(Context &&context) noexcept = default;

        Context &operator=(const Context &context) = delete;

        Context &operator=(Context &&context) = default;

        explicit Context(SdlWindow &window, int index, Uint32 flags) : Context()
        { mRenderer.reset(SDL_CreateRenderer(window.get(), index, flags)); }

        /// Test for a valid Context
        explicit operator bool() const noexcept { return mRenderer.operator bool(); }

        /// Get an opaque pointer for API calls.
        [[nodiscard]] auto get() const { return mRenderer.get(); }

        /// Set the draw blend mode.
        void setDrawBlendMode(SDL_BlendMode blendMode) {
            SDL_SetRenderDrawBlendMode(mRenderer.get(), blendMode);
        }

        /**
         * @brief Copy source Texture to destination Texture and set the BlendMode on the destination Texture.
         * @details The function uses RenderTargetGuard to temporarily set the render Target to the destination,
         * temporarily sets the source blend mode to copyBlendMode, calls SDL_RenderCopy to copy the texture,
         * sets the BlendMode on the destination texture to destinationBlendMode and returns the source blend
         * mode to its original value.
         * <p/>
         * If the
         * @param source The source Texture
         * @param destination The destination Texture
         * @param copyBlendMode The blend mode used while copying source to destination.
         * @param destinationBlendMode The final blend mode set on the destination.
         * @return RoseErrorCode
         */
        [[maybe_unused]] bool copyTexture(Texture &source, Texture &destination,
                                          SDL_BlendMode copyBlendMode = SDL_BLENDMODE_NONE,
                                          SDL_BlendMode destinationBlendMode = SDL_BLENDMODE_BLEND);

        /// Prepare for the start of a rendering iteration.
        int renderClear() { return SDL_RenderClear(mRenderer.get()); }

        /// Complete a rendering iteration.
        void renderPresent() { SDL_RenderPresent(mRenderer.get()); }

        /**
         * @brief Copy a Texture to the current render target using the size of the Texture and the size of the
         * target.
         * @param texture The texture to copy.
         * @return The return status of the API call.
         */
        bool renderCopy(const Texture &texture);

        /**
         * @brief Copy a Texture to the current render target using the size of the Texture and specified
         * destination rectangle.
         * @param texture The source Texture.
         * @param dst The destination Rectangle.
         * @return The return status of the API call.
         */
        bool renderCopy(const Texture &texture, Rectangle<int> dst);

        /**
         * @brief Copy part of the Texture specified by the source Rectangle to the render target location
         * specified by the destination Rectangle.
         * @param texture The Texture.
         * @param src The source Rectangle.
         * @param dst The destination Rectangle.
         * @return The return status of API call.
         */
        [[maybe_unused]] bool renderCopy(const Texture &texture, Rectangle<int> src, Rectangle<int> dst);

        /**
         * @brief Render with extensions.
         * @param src Source Rectangle.
         * @param dst Destination Rectangle.
         * @param angle Rotation Angle.
         * @param renderFlip Flip
         * @param point Rotation Position if specified.
         * @return Status code returned by SDL_RenderCopyEx()
         */
        [[maybe_unused]] bool
        renderCopyEx(Texture &texture, Rectangle<int> src, Rectangle<int> dst, double angle, RenderFlip renderFlip,
                     std::optional<Position<int>> point = std::nullopt) const;

        /**
         * @brief Set the drawing color used for drawing Rectangles, lines and clearing.
         * @param color The new drawing Color.
         * @return Status code returned by the API.
         */
        bool setDrawColor(color::RGBA color);

        bool setDrawColor(SDL_Color c);

        /**
         * @brief Determine if a clip rectangle is enabled on the Context renderer.
         * @return
         */
        [[nodiscard]] bool isClipEnabled() const {
            return SDL_RenderIsClipEnabled(get());
        }

        /**
         * @brief Set the drawing color used for drawing Rectangles, lines and clearing.
         * @param color The new drawing Color.
         * @return Status code returned by the API.
         */
        [[maybe_unused]] int setDrawColor(color::HSVA color) {
            return setDrawColor(color.toRGBA());
        }

        /**
         * @brief Render a filled Rectangle.
         * @param rect The Rectangle
         * @param color The fill color.
         * @return true if successful, false on error, see rose::ErrorStatus.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        bool fillRect(Rectangle<T> rect, SDL_Color color) {
            DrawColorGuard drawColorGuard{*this, color};
            if constexpr (std::is_integral_v<T>) {
                SDL_Rect r{rect.x, rect.y, rect.w, rect.h};
                if (SDL_RenderFillRect(get(), &r))
                    return RoseError(ErrorStatus::SDL_ERROR);
            } else {
                SDL_FRect r{rect.x, rect.y, rect.w, rect.h};
                if (SDL_RenderFillRectF(get(), &r))
                    return RoseError(ErrorStatus::SDL_ERROR);
            }
            return true;
        }


        /**
         * @brief Render a filled Rectangle.
         * @param rect The Rectangle
         * @param color The fill color.
         * @return true if successful, false on error, see rose::ErrorStatus.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        bool fillRect(Rectangle<T> rect, color::RGBA color) {
            DrawColorGuard drawColorGuard{*this, color};
            if constexpr (std::is_integral_v<T>) {
                SDL_Rect r{rect.x, rect.y, rect.w, rect.h};
                if (SDL_RenderFillRect(get(), &r))
                    return RoseError(ErrorStatus::SDL_ERROR);
            } else {
                SDL_FRect r{rect.x, rect.y, rect.w, rect.h};
                if (SDL_RenderFillRectF(get(), &r))
                    return RoseError(ErrorStatus::SDL_ERROR);
            }
            return true;
        }

        /**
         * @brief Render a pixel.
         * @param p The location of the pixel.
         * @return The status return from the SDL API.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        [[maybe_unused]] bool drawPoint(const Position<T> &p, const color::RGBA &color) {
            gm::DrawColorGuard drawColorGuard{*this, color};
            if constexpr (std::is_integral_v<T>) {
                if (SDL_RenderDrawPoint(get(), p.x, p.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            } else {
                if (SDL_RederDrawPointF(get(), p.x, p.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            }
            return true;
        }

        /**
         * @brief Render a pixel.
         * @param p The location of the pixel.
         * @return The status return from the SDL API.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        [[maybe_unused]] bool drawPoint(const Position<T> &p) {
            if constexpr (std::is_integral_v<T>) {
                if (SDL_RenderDrawPoint(get(), p.x, p.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            } else {
                if (SDL_RederDrawPointF(get(), p.x, p.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            }
            return true;
        }

        /**
         * @brief Draw a line.
         * @tparam T The type of the position data.
         * @param p0 The starting point of the line.
         * @param p1 The ending point of the line
         * @return true if successful, false on error see rose::ErrorStatus.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        [[maybe_unused]] bool drawLine(const Position<T> &p0, const Position<T> &p1) const {
            if constexpr(std::is_integral_v<T>) {
                if (SDL_RenderDrawLine(get(), p0.x, p0.y, p1.x, p1.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            } else {
                if (SDL_RenderDrawLineF(get(), p0.x, p0.y, p1.x, p1.y))
                    return RoseError(ErrorStatus::SDL_ERROR);
            }
            return true;
        }

        /**
         * @brief Draw a line.
         * @tparam T The type of the position data.
         * @param line The line.
         * @return true if successful, false on error see rose::ErrorStatus.
         */
        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        bool drawLine(const Line<T> &line) const {
            return drawLine(line.p0, line.p1);
        }
    };
}

