/**
 * @file DrawColorGuard.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-20
 */

#pragma once

#include <SDL.h>
#include "Color.h"

namespace rose::gm {
    class Context;

    /**
     * @class DrawColorGuard
     * @brief Store the current draw color replacing it with a new draw color. When the object is
     * destroyed (by going out of scope) the old draw color is set.
     */
    class DrawColorGuard {
    protected:
        Context &mContext;      ///< The renderer to which the draw colors are set.
        SDL_Color mOldColor{};  ///< The old draw color
        bool mStatus{false};    ///< The status of the last SDL operation.

    public:
        DrawColorGuard() = delete;  ///< Deleted default constructor.
        DrawColorGuard(const DrawColorGuard &) = delete; ///< Deleted copy constructor.
        DrawColorGuard(DrawColorGuard &&) = default;

        DrawColorGuard &operator=(const DrawColorGuard &) = delete;

        /**
         * @brief Set the old clip rectangle back on the renderer when destroyed.
         */
        ~DrawColorGuard() noexcept(false);

        /**
         * @brief Construct DrawColorGuard with and SDL_Color.
         * @param renderer The renderer to set the color on.
         * @param color The SDL_Color.
         */
        DrawColorGuard(Context &context, SDL_Color color);

        /**
         * @brief Construct DrawColorGuard with a rose::Color.
         * @param renderer The renderer to set the color on.
         * @param color The rose::Color.
         */
        DrawColorGuard(Context &context, color::RGBA color);

        /**
         * @brief Test the validity of the DrawColorGuard.
         * @return False if the last SDL operation return an error.
         */
        [[nodiscard]] explicit operator bool() const noexcept { return mStatus; }

        /**
         * @brief Set the draw Color on the renderer without pushing the old color on the stack.
         * @param color The SDL_Color to set.
         * @return The SDL2 API return status code.
         */
        bool setDrawColor(SDL_Color color);

        /**
         * @brief Set the draw Color on the renderer without pushing the old color on the stack.
         * @param color The Color to set.
         * @return The SDL2 API return status code.
         */
        [[maybe_unused]] int setDrawColor(color::RGBA color) {
            return setDrawColor(color.toSdlColor());
        }
    };
}

