/**
 * @file Event.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-04
 */

#pragma once

#include <SDL_events.h>
#include <string>
#include "Sdl++/Metrics.h"

namespace rose {

    /**
     * @class Event
     * @brief
     */
    class Event {
    public:
        enum EventType : uint32_t {
            WINDOW = SDL_WINDOWEVENT,
            KEYDOWN = SDL_KEYDOWN,
            KEYUP = SDL_KEYUP,
            TEXT_EDITING = SDL_TEXTEDITING,
            TEXT_INPUT = SDL_TEXTINPUT,
            MOUSE_MOTION = SDL_MOUSEMOTION,
            MOUSE_BUTTON_DOWN = SDL_MOUSEBUTTONDOWN,
            MOUSE_BUTTON_UP = SDL_MOUSEBUTTONUP,
            MOUSEWHEEL = SDL_MOUSEWHEEL,
            QUIT = SDL_QUIT,
            USER = SDL_USEREVENT,
            FINGER_MOTION = SDL_FINGERMOTION,
            FINGER_DOWN = SDL_FINGERDOWN,
            FINGER_UP = SDL_FINGERUP,
            MULTI_GESTURE = SDL_MULTIGESTURE,
            DOLLAR_GESTURE = SDL_DOLLARGESTURE,
            DOLLAR_RECORD = SDL_DOLLARRECORD,
            DROP_FILE = SDL_DROPFILE,
            DROP_TEXT = SDL_DROPTEXT,
            DROP_BEGIN = SDL_DROPBEGIN,
            DROP_COMPLETE = SDL_DROPCOMPLETE,
            NONE,
            MOUSE_LEFT,
            MOUSE_ENTER,
        };

        EventType eventType{EventType::NONE};
        uint32_t timeStamp{}, x{}, y{};
        bool locatedEvent{false};
        [[nodiscard]] Position<int> eventPosition() const { return Position<int>{x, y}; }

        Event() noexcept = default;

        explicit Event(const uint32_t type, uint32_t time_stamp) noexcept:
                eventType(static_cast<EventType>(type)), timeStamp(time_stamp) {}

        explicit Event(const EventType type, uint32_t time_stamp) noexcept:
                eventType(type), timeStamp(time_stamp) {}

        ~Event() = default;
    };

    class WindowIdEvent : public Event {
    public:
        uint32_t windowId{};

        WindowIdEvent(uint32_t type, uint32_t timestamp, uint32_t window_id) noexcept:
                Event(static_cast<EventType>(type), timestamp), windowId(window_id) {}
    };

    class WindowEvent : public WindowIdEvent {
    public:
        uint32_t windowId{};
        uint8_t event{};
        int32_t data1{};
        int32_t data2{};

        explicit WindowEvent(const SDL_WindowEvent &windowEvent) noexcept:
                WindowIdEvent(windowEvent.type, windowEvent.timestamp, windowEvent.windowID),
                event(windowEvent.event), data1(windowEvent.data1), data2(windowEvent.data2) {}
    };

    class KeyboardEvent : public WindowIdEvent {
    public:
        uint32_t windowId{};
        uint8_t state{};
        uint8_t repeat{};
        SDL_Keysym key_symbol{};

        explicit KeyboardEvent(const SDL_KeyboardEvent &keyboardEvent) noexcept:
                WindowIdEvent(keyboardEvent.type, keyboardEvent.timestamp, keyboardEvent.windowID),
                state(keyboardEvent.state), repeat(keyboardEvent.repeat), key_symbol(keyboardEvent.keysym) {}
    };

    class TextInputEvent : public WindowIdEvent {
    public:
        std::string text{};

        TextInputEvent(uint32_t type, uint32_t timestamp, uint32_t window_id, const char *c_str) noexcept:
                WindowIdEvent(type, timestamp, window_id), text(c_str) {}

        explicit TextInputEvent(const SDL_TextInputEvent &textInputEvent) noexcept:
                WindowIdEvent(textInputEvent.type, textInputEvent.timestamp, textInputEvent.windowID),
                text(textInputEvent.text) {}
    };

    class TextEditingEvent : public TextInputEvent {
    public:
        int32_t start{}, length{};

        explicit TextEditingEvent(const SDL_TextEditingEvent &textEditingEvent) noexcept:
                TextInputEvent(textEditingEvent.type, textEditingEvent.timestamp, textEditingEvent.windowID,
                               textEditingEvent.text),
                start(textEditingEvent.start), length(textEditingEvent.length) {}
    };

    class MouseEvent : public WindowIdEvent {
    public:
        uint32_t which;

        MouseEvent(uint32_t type, uint32_t timestamp, uint32_t window_id, uint32_t whichMouse, int32_t X,
                   int32_t Y) noexcept:
                WindowIdEvent(type, timestamp, window_id), which(whichMouse) {
            locatedEvent = true;
            x = X;
            y = Y;
        }
    };

    class MouseMotionEvent : public MouseEvent {
    public:
        uint32_t state;
        int32_t xRel, yRel;

        explicit MouseMotionEvent(const SDL_MouseMotionEvent &mouseMotionEvent) noexcept:
                MouseEvent(mouseMotionEvent.type, mouseMotionEvent.timestamp, mouseMotionEvent.windowID,
                           mouseMotionEvent.which, mouseMotionEvent.x, mouseMotionEvent.y),
                state(mouseMotionEvent.state), xRel(mouseMotionEvent.xrel), yRel(mouseMotionEvent.yrel) {}
    };

    class MouseButtonEvent : public MouseEvent {
    public:
        uint8_t state, button, clicks;

        explicit MouseButtonEvent(const SDL_MouseButtonEvent &mouseMotionEvent) noexcept:
                MouseEvent(mouseMotionEvent.type, mouseMotionEvent.timestamp, mouseMotionEvent.windowID,
                           mouseMotionEvent.which, mouseMotionEvent.x, mouseMotionEvent.y),
                state(mouseMotionEvent.state), button(mouseMotionEvent.button), clicks(mouseMotionEvent.clicks) {}
    };

    class MouseWheelEvent : public MouseEvent {
    public:
        int32_t dX, dY;

        MouseWheelEvent(const SDL_MouseWheelEvent &mouseWheelEvent, const Position<int>& mousePosition) noexcept:
                MouseEvent(mouseWheelEvent.type, mouseWheelEvent.timestamp, mouseWheelEvent.windowID,
                           mouseWheelEvent.which, mousePosition.x, mousePosition.y) {
            dX = mouseWheelEvent.x;
            dY = mouseWheelEvent.y;
            if (mouseWheelEvent.direction == SDL_MOUSEWHEEL_FLIPPED) {
                x *= -1;
                y *= -1;
            }
        }
    };
}

