/**
 * @file Graph.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-28
 */

#pragma once

#include <concepts>
#include "SDL++.h"
#include "Color.h"
#include "Surface.h"

namespace rose {

    struct GraphPen {
        color::RGBA drawColor{color::RGBA::OpaqueWhite};
        int thickness{1};
        bool antiAlias{true};
    };

    /**
     * @class Graph
     * @brief A base class to help create a texture containing a data graph.
     */
    class Graph {
    protected:
        Rectangle<int> mRectangle{};
        Margin<int> plotMargin{};
        Rectangle<int> plotRectangle{};
        Rectangle<float> plotCoordinates{0., 0., 1., 1.};
        color::RGBA plotBackground{color::RGBA::TransparentBlack};
        gm::Surface plotSurface;

    public:
        Graph() = delete;

        explicit Graph(Rectangle<int> rectangle) : mRectangle(rectangle), plotSurface(rectangle.size()) {
            clear();
        }

        ~Graph() = default;

        template<typename T>
        requires std::integral<T>
        void setMargin(const Margin<T> margin) {
            plotMargin = margin;
        }

        template<typename T>
        requires std::floating_point<T>
        void setMargin(const Margin<T> margin) {
            plotMargin.top = util::roundToInt(margin.top * mRectangle.h);
            plotMargin.bot = util::roundToInt(margin.bot * mRectangle.h);
            plotMargin.left = util::roundToInt(margin.left * mRectangle.h);
            plotMargin.right = util::roundToInt(margin.right * mRectangle.h);
        }

        /**
         * @brief Set the SDL_BlendMode of the plot surface.
         * @param blendMode The blend mode.
         * @return true on success, false on failrue. See ros::ErrorStatus.
         */
        bool setBlendMode(SDL_BlendMode blendMode) {
            return plotSurface.setBlendMode(blendMode);
        }

        template<typename T>
        requires std::floating_point<T>
        Position<int> convertFromCoord(const Position<T> &position) {
            Position<int> p{};

            p.x = util::roundToInt(((position.x - plotCoordinates.x) / plotCoordinates.w) * plotRectangle.w) +
                  plotRectangle.x;
            p.y = util::roundToInt(((position.y - plotCoordinates.y) / plotCoordinates.h) * plotRectangle.h) +
                  plotRectangle.y;
            return p;
        }

        /**
         * @brief Clear the surface and set up the parametric information.
         */
        void clear() {
            if (plotSurface) {
                mRectangle = PositionZero;
                mRectangle = plotSurface.size();
                plotRectangle = plotMargin.position();
                plotRectangle = mRectangle.size() - plotMargin.size();
                plotSurface.fillRectangle(mRectangle, plotBackground);
            }
        }

        bool drawLineIntegral(const GraphPen &pen, Line<int> line);

        template<typename T>
        requires std::integral<T> || std::floating_point<T>
        bool drawLine(const GraphPen &pen, Line <T> line) {
            if constexpr(std::integral<T>) {
                return drawLineIntegral(pen, line);
            } else {
                Line<int> convertedLine(convertFromCoord(line.p0), convertFromCoord(line.p1));
                return drawLineIntegral(pen, convertedLine);
            }
        }
    };
}

