/**
 * @file ElapsedTimeGuard.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-17
 */

#pragma once

#include <chrono>
#include <iostream>
#include <string>

namespace rose {

    /**
     * @class SingletonGuard
     * @brief Gets a reference to a singleton object and makes it available through pointer semantics.
     * @details The Singleton must have an initializer function get() that returns a referance to the
     * Singleton object.
     * @tparam Singleton The type of the Singleton.
     */
    template<class Singleton>
    class SingletonGuard {
    private:

        Singleton& mSingleton;

    public:
        SingletonGuard() : mSingleton(Singleton::get()) {}

        ~SingletonGuard() = default;

        SingletonGuard(const SingletonGuard&) = delete;
        SingletonGuard(SingletonGuard&&) = delete;
        SingletonGuard& operator = (const SingletonGuard&) = delete;
        SingletonGuard& operator = (SingletonGuard&&) = delete;

        Singleton* operator -> () { return &mSingleton; }
        const Singleton* operator -> () const { return &mSingleton; }
    };


    /**
     * @struct ElapsedTimeGuardUnits
     * @brief Deduce units for the ElapsedTimeGuard.
     * @tparam D
     */
    template<typename D>
    struct ElapsedTimeGuardUnits {
        static constexpr std::string_view value{};
    };

    template<typename D>
    static constexpr std::string_view ElapsedTimeGuardUnitsUnits_v = ElapsedTimeGuardUnits<D>::value;

    template<>
    struct ElapsedTimeGuardUnits<std::chrono::nanoseconds> {
        static constexpr std::string_view value{" ns"};
    };

    template<>
    struct ElapsedTimeGuardUnits<std::chrono::microseconds> {
        static constexpr std::string_view value{" us"};
    };

    template<>
    struct ElapsedTimeGuardUnits<std::chrono::milliseconds> {
        static constexpr std::string_view value{" ms"};
    };

    template<>
    struct ElapsedTimeGuardUnits<std::chrono::seconds> {
        static constexpr std::string_view value{" s"};
    };

    /**
     * @class ElapsedTimeGuard
     * @brief Measures and
     * @brief
     */
    template<typename Duration = std::chrono::milliseconds>
    class ElapsedTimeGuard {
    public:
        std::chrono::system_clock::time_point start, stop;
        std::string message;
        std::ostream& ostream;

        template<class S>
        ElapsedTimeGuard(S message, std::ostream& ostream) : message(message), ostream(ostream) {
            static_assert(std::is_constructible_v<std::string, S>, "Must be able to construct std::string from S.");
            start = std::chrono::system_clock::now();
        }

        template<class S>
        explicit ElapsedTimeGuard(S message) : ElapsedTimeGuard(message, std::cout) {}

        ~ElapsedTimeGuard() {
            stop = std::chrono::system_clock::now();
            ostream << message << ' '
                      << std::chrono::duration_cast<Duration>(stop - start).count()
                      << ElapsedTimeGuardUnitsUnits_v<Duration> << '\n';
        }
    };
}

