/**
 * @file Position.b
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-27
 */

#pragma once

#include <compare>
#include <concepts>
#include <type_traits>
#include "SDL++.h"

namespace rose {
    /**
     * @struct Position
     * @brief A position in integer (x, y) co-ordinates.
     * @details For screen objects (0, 0) is the top left corner, x increases to the right, y increases toward
     * the bottom.
     */
    template<typename T>
    requires std::integral<T> || std::floating_point<T>
    struct Size;

    template<typename T>
    requires std::integral<T> || std::floating_point<T>
    struct Position {
        T x{0}, y{0};

        constexpr Position() noexcept = default;
        constexpr Position(const Position &p) = default;
        constexpr Position(Position &&p)  noexcept = default;
        constexpr Position& operator=(const Position &p) = default;
        constexpr Position& operator=(Position &&p)  noexcept = default;

        explicit Position(T p) noexcept : x(p), y(p) {}

        template<class O>
        constexpr Position(O X, O Y) : x(X), y(Y) {}

        /**
         * @brief Three way comparison of two Positions.
         * @details If corresponding coordinates of this and other are equivalent, then this is equivalent to other.
         * If one co-ordinate of this is less than its correspondent of other and the orthogonal
         * is less than or equal to its correspondent then this is less than other. If one co-ordinate of this is
         * greater than its correspondent of other and the orthogonal is greater than or equal to its
         * correspondent then this is greater than other. Otherwise the this and other are unordered.
         * @param other The other Position
         * @return A std::partial_ordering value.
         */
        constexpr std::partial_ordering operator <=>(const Position& other) const noexcept {
            auto xo = x <=> other.x;
            auto yo = y <=> other.y;

            if ((xo < 0 && yo <= 0) || (xo <= 0 && yo < 0))
                return std::partial_ordering::less;
            if ((xo > 0 && yo >= 0) || (xo >= 0 && yo > 0))
                return std::partial_ordering::greater;
            if (xo == 0 && yo == 0)
                return std::partial_ordering::equivalent;

            return std::partial_ordering::unordered;
        }

        constexpr bool operator == (const Position& other) const noexcept {
            return operator<=>(other) == 0;
        }

        constexpr bool operator != (const Position& other) const noexcept {
            return operator<=>(other) != 0;
        }

        /// Add two positions together.
        [[nodiscard]] constexpr Position operator+(const Position<T> &p) const noexcept {
            return Position<T>(this->x+p.x,this->y+p.y);
        }

        /// Subtract position other from this.
        [[nodiscard]] constexpr Position operator-(const Position<T> &p) const noexcept {
            return Position<T>(this->x-p.x,this->y-p.y);
        }

        [[nodiscard]] T& primary(Orientation orientation) {
            return orientation == Orientation::Horizontal ? x : y;
        }

        [[nodiscard]] T& secondary(Orientation orientation) {
            return orientation == Orientation::Vertical ? x : y;
        }

        [[nodiscard]] const T& primary(Orientation orientation) const {
            return orientation == Orientation::Horizontal ? x : y;
        }

        [[nodiscard]] const T& secondary(Orientation orientation) const {
            return orientation == Orientation::Vertical ? x : y;
        }

        std::ostream& printOn(std::ostream& strm) const {
            return strm << '[' << x << ',' << y << ']';
        }

        template<typename C>
        Position<C> convertTo(const Position<T>& p) {
            return Position<C>(util::roundTo<C>(x), util::roundTo<C>(y));
        }
    };

    static constexpr Position<int> PositionZero;
    static constexpr Position<int> UndefinedPosition{SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED};
    static constexpr Position<int> PositionCentered{ SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED};

    template<typename T>
    struct Line {
        Position<T> p0{}, p1{};

        constexpr Line() noexcept = default;
        ~Line() noexcept = default;
        constexpr Line(const Line &) noexcept = default;
        constexpr Line(Line &&) noexcept = default;
        constexpr Line& operator = (const Line &) noexcept = default;
        constexpr Line(const Position<T>& p0, const Position<T>& p1) noexcept : p0(p0), p1(p1) {}
        constexpr Line(Position<T>&& p0, Position<T>&& p1) noexcept : p0(p0), p1(p1) {}
        constexpr Line(T x0, T y0, T x1, T y1) noexcept : p0(x0,y0), p1(x1,y1) {}

        constexpr std::tuple<T,T,T,T> to_tuple() const noexcept {
            return std::make_tuple(p0.x, p0.y, p1.x, p1.y);
        }

        constexpr std::pair<Position<T>,Position<T>> to_pair() const noexcept {
            return std::make_pair(p0, p1);
        }
    };
}
