/**
 * @file Rectangle.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-27
 */

#pragma once

#include <tuple>
#include "Size.h"
#include "Position.h"

namespace rose {
    /**
     * @struct Rectangle
     * @brief A composite of a Position and a Size.
     */
    template<typename T=int>
    struct Rectangle {
        T x{}, y{}, w{}, h{};

        constexpr Rectangle() noexcept = default;
        constexpr Rectangle(T X, T Y, T W, T H) noexcept : x(X), y(Y), w(W), h(H) {}
        constexpr Rectangle(const Position<T> &p, const Size<T> &s) noexcept : Rectangle(p.x, p.y, s.w, s.h) {}
        constexpr Rectangle(const Rectangle &rectangle) = default;
        constexpr Rectangle(Rectangle &&rectangle)  noexcept = default;
        constexpr Rectangle& operator=(const Rectangle &rectangle) = default;
        constexpr Rectangle& operator=(Rectangle &&rectangle)  noexcept = default;

        constexpr explicit Rectangle(const SDL_Rect& rect) : x(rect.x), y(rect.y), w(rect.w), h(rect.h) {
            static_assert(std::is_integral_v<T>, "Conversion only supported for integer Rectangles");
        }

        constexpr Rectangle& operator=(const Size<T> &s) {
            Rectangle<T>::w = s.w;
            Rectangle<T>::h = s.h;
            return *this;
        }

        constexpr Rectangle& operator=(const Position<T> &p) {
            Rectangle<T>::x = p.x;
            Rectangle<T>::y = p.y;
            return *this;
        }

        constexpr Rectangle& operator=(const SDL_Rect& r) {
            static_assert(std::is_integral_v<T>, "Assignment only supported for integer Rectangles");
            x = r.x;
            y = r.y;
            w = r.w;
            h = r.h;
            return *this;
        }

        [[nodiscard]] constexpr SDL_Rect sdlRect() const noexcept {
            static_assert(std::is_integral_v<T>, "Conversion only supports integer Rectanbles.");
            SDL_Rect result{x, y, w, h};
            return result;
        }

        /// Add a Position to a Rectangle.
        [[nodiscard]] constexpr Rectangle operator + (const Position<T> &p) const noexcept {
            return Rectangle{this->x + p.x, this->y + p.y, this->w, this->h};
        }

        /// Subtract a Position from a Rectangle.
        [[nodiscard]] constexpr Rectangle operator - (const Position<T> &p) const noexcept {
            return Rectangle{this->x - p.x, this->y - p.y, this->w, this->h};
        }

        /// Add a Size to a Rectangle.
        [[nodiscard]] constexpr Rectangle operator + (const Size<T> &s) const noexcept {
            return Rectangle{ this->x, this->y, this->w + s.w, this->h + s.h};
        }

        /// Subtract a Size from a Rectangle.
        [[nodiscard]] constexpr Rectangle operator - (const Size<T> &s) const noexcept {
            return Rectangle{ this->x, this->y, this->w - s.w, this->h - s.h};
        }

        /// Get a Position from a Rectangle.
        [[nodiscard]] Position<T> position() const noexcept {
            return Position<T>{this->x,this->y};
        }

        /// Get a Size from a Rectangle
        [[nodiscard]] Size<T> size() const noexcept {
            return Size{this->w,this->h};
        }

        /// Get the Positions of top-left and bottom-right corners defined by the Rectangle
        [[nodiscard]] std::pair<Position<T>,Position<T>> primeCorners() const noexcept {
            return std::make_pair(Position<T>{this->x, this->y}, Position < T > {this->x + this->w, this->y + this->h});
        }

        /// Get the Positions of the top-right and bottom-left corners defined by the Rectangle
        [[nodiscard]] std::pair<Position<T>,Position<T>> crossCorners() const noexcept {
            return std::make_pair(Position<T>{this->x + this->w, this->y}, Position < T > {this->x, this->y + this->h});
        }

        /// Get the Positions of all four corners defined by the Rectangle, top to bottom, left to right.
        [[nodiscard]] std::tuple<Position<T>,Position<T>,Position<T>,Position<int>> corners() const noexcept {
            auto [tl,br] = primeCorners();
            auto [bl,tr] = crossCorners();
            return std::make_tuple(tl,tr,bl,br);
        }

        /// Determine if a Rectangle containsRelative a Position.
        [[nodiscard]] constexpr bool contains(Position<T> pos) const noexcept {
            return pos.x >= this->x && pos.x < this->x + this->w && pos.y >= this->y && pos.y < this->y + this->h;
        }

        /**
         * @brief Determine if there is no overlap between two Rectangle objects.
         * @param o The other Rectangle.
         * @return True if there is no overlap of this Rectangle and the other Rectangle.
         */
        [[nodiscard]] constexpr bool noOverlap(const Rectangle& o) const noexcept {
            return this->x > o.x + o.w || o.x > this->x + this->w || this->y > o.y + o.h || o.y > this->y + this->h;
        }

        /**
         * @brief Determine if there is overlap between to Rectangle objects by inverting noOverlap().
         * @param o The other Rectangle.
         * @return True if there is overlap of this Rectangle and the other Rectangle.
         */
        [[nodiscard]] constexpr bool overlap(const Rectangle& o) const noexcept {
            return !noOverlap(o);
        }

        [[nodiscard]] Rectangle intersection(const Rectangle &o) {
            // gives bottom-left point
            // of intersection rectangle

            auto x5 = std::max(x, o.x);
            auto y5 = std::max(y, o.y);

            // gives top-right point
            // of intersection rectangle
            auto x6 = std::min(x+w, o.x+o.w);
            auto y6 = std::min(y+h, o.y+o.h);

            // no intersection
            if (x5 > x6 || y5 > y6) {
                return Rectangle{0,0,0,0};
            }

            return Rectangle{x5, y5, x6 - x5, y6 - y5};
        }

        [[nodiscard]] std::optional<Line<T>> intersectLine(const Line<T>& line) const noexcept {
            static_assert(std::is_integral_v<T>, "Method only valid for integer Rectangles nad Lines");
            auto r = sdlRect();
            auto res = line;
            if (SDL_IntersectRectAndLine(&r, &res.p0.x, &res.p0.y, &res.p1.x, &res.p1.y)) {
                return res;
            }
            return std::nullopt;
        }

        [[nodiscard]] std::optional<Rectangle<int>> intersectRect(const Rectangle<int>& b) const noexcept {
            SDL_Rect result;
            auto A = sdlRect();
            auto B = b.sdlRect();
            if (SDL_IntersectRect(&A, &B, &result))
                return Rectangle<int>(result);
            return std::nullopt;
        }
    };
}
