/**
 * @file RenderTargetGuard.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-20
 */

#pragma once

#include <SDL.h>
#include "Texture.h"

namespace rose::gm {

    class Context;

    /**
     * @class RenderTargetGuard
     * @details Setting a render target will remove any set clip rectangle silently making any current
     * ClipRectangleGuard ineffective. This is an undocumented feature of the SDL library.
     * @brief Store the current render target replacing it with a new render target. When the object is
     * destroyed (by going out of scope) the old render target is restored.
     */
    class RenderTargetGuard {
    protected:
        Context &mContext;                      ///< The Context being guarded
        bool popped{false};                     ///< True if the target guard has already been popped off the stack.
        SDL_Texture *mLastTexture{nullptr};     ///< Save the current render target here.
        bool status{true};                      ///< The return status from the last SDL API call.

    public:
        RenderTargetGuard() = delete;   ///< Deleted default constructor.
        RenderTargetGuard(const RenderTargetGuard &) = delete;   ///< Deleted copy constructor.

        /**
         * @brief Set the old render target back on the renderer when destroyed.
         */
        ~RenderTargetGuard() noexcept(false);

        /**
         * @brief Test the status of the RenderTargetGuard.
         * @details The status is good (true is returned) if the status value returned from the last
         * operation on the Context object returned 0.
         * @return True if the last operation succeeded.
         */
        explicit operator bool() const noexcept { return status == 0; }

        /**
         * @brief Constructor
         * @details Setting a render target will clear the clip rectangle in the context renderer. If the
         * texture was not created with SDL_TEXTUREACCESS_TARGET this method will set the status to false
         * and set the rose::ErrorStatus appropriately.
         * @param renderer The renderer which render target will be managed.
         * @param texture The texture which will become the new render target.
         */
        RenderTargetGuard(Context &context, Texture &texture);

        /**
         * @brief Set a new render target without pushing it on the stack.
         * @details This may be used when a number of render target changes are needed in a context block.
         * A RenderTargetGuard is created, calls to setRenderTarget are used to manipulate the render target.
         * When the RenderTargetGuard goes out of scope the original render target is restored. If the
         * texture was not created with SDL_TEXTUREACCESS_TARGET this method will set the status to false
         * and set the rose::ErrorStatus appropriately.
         * @param texture
         * @return true if the render target is set, false otherwise and ErrorStatus is set.
         */
        [[maybe_unused]] bool setRenderTarget(Texture &texture);
    };
}

