/**
 * @file Resources.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-26
 */

#pragma once

#include <array>
#include <filesystem>
#include <iostream>
#include <soci/soci.h>
#include <sqlite3/soci-sqlite3.h>
#include "Sdl++/Metrics.h"
#include "Theme.h"
#include "Sdl++/XDGFilePaths.h"

namespace rose {

    class Resources {
    public:
        using database_signature = std::array<std::string_view, 3>;

        enum class Status : std::size_t {
            OK,
            DB_ERROR,
            DB_INFO_WRONG,
            NOT_INITIALIZED,
        };

        static constexpr std::array<std::string_view, 4> ErrorMessage =
                {{
                         "Ok.",
                         "Sqlite database error.",
                         "Database information record does not match requested database.",
                         "Database not initialized, constructor probably not called.",
                 }};
    protected:
        bool mValid{false};
        Status mStatus{Status::NOT_INITIALIZED};
        std::filesystem::path mPath{};
        std::string mType{};
        std::string mComponent{};
        std::string mComponentName{};
        std::array<int, 4> mVersion{};

    public:
        Resources() = delete;

        ~Resources() = default;

        Resources(const std::filesystem::path &dbPath, const Resources::database_signature &signature) {
            mPath = dbPath;
            try {
                using namespace soci;
                soci::session sql(sqlite3, dbPath.string());
                row r;
                sql << "SELECT type,component,component_name FROM dbinfo LIMIT 1", into(r);
                if (sql.got_data() && r.get_indicator(0) == i_ok && r.get_indicator(1) == i_ok &&
                    r.get_indicator(2) == i_ok) {
                    mType = r.get<std::string>(0);
                    mComponent = r.get<std::string>(1);
                    mComponentName = r.get<std::string>(2);
                    if (mType == signature[0] && mComponent == signature[1] && mComponentName == signature[2]) {
                        mValid = true;
                        mStatus = Status::OK;
                    }
                }
            } catch (const std::exception &e) {
                std::cerr << e.what() << '\n';
                mValid = false;
                mStatus = Status::DB_ERROR;
            }
        }

        auto openSession() {
            return soci::session(soci::sqlite3, mPath.string());
        }

        bool operator==(const std::filesystem::path &path) const {
            return mPath == path;
        }

        explicit operator bool() const {
            return mValid;
        }

        [[nodiscard]] constexpr const std::string_view error() const {
            return (ErrorMessage[static_cast<std::size_t>(mStatus)]);
        }
    };

    class ResourceSet {
    protected:
        std::vector<Resources> mResources{};

        static constexpr std::string_view intTable = "int";
        static constexpr std::string_view intPairTable = "intpair";
        static constexpr std::string_view intQuadTable = "intquad";

    public:
        template<size_t N>
        void addResources(XDGFilePaths &filePaths, const std::array<XDGFilePaths::XDG_Name, N> &resourceNames,
                          const Resources::database_signature &signature, const std::string &name) {
            for (auto &configName: resourceNames) {
                auto[found, path] = filePaths.template findFilePath(configName, name);
                auto foundPath = path;
                if (found) {
                    if (std::find_if(mResources.begin(), mResources.end(),
                                     [&foundPath](const Resources &res) {
                                         return res == foundPath;
                                     }) == mResources.end()) {
                        Resources resources{path, signature};
                        if (resources)
                            mResources.emplace_back(resources);
                    }
                }
            }
        }

        ResourceSet() = default;

        ~ResourceSet() = default;

        [[nodiscard]] explicit operator bool () const noexcept {
            return !mResources.empty();
        }

        auto begin() { return mResources.begin(); }

        auto end() { return mResources.end(); }

        auto front() { return mResources.front(); }

//        template<class S>
//        std::optional<int> getValue(soci::session &sql, S classname, S parameter) {
//            using namespace soci;
//            int value;
//            sql << "SELECT value FROM " << intTable << " WHERE class = \"" << classname << "\" AND parameter = \""
//                << parameter << '"', into(value);
//            if (sql.got_data())
//                return value;
//            return std::nullopt;
//        }

        template<typename T>
        std::optional<T>
        getValue(soci::session &sql, const std::string_view classname, const std::string_view parameter) {
            static_assert(std::is_same_v<T,bool>, "No resource handler for type.");
        }
    };

    template<>
    inline std::optional<bool>
    ResourceSet::getValue<bool>(soci::session &sql, const std::string_view classname,
                                const std::string_view parameter) {
        using namespace soci;
        int a;
        sql << "SELECT value FROM " << intTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(a);
        if (sql.got_data())
            return a != 0;

        return std::nullopt;
    }

    template<>
    inline std::optional<int>
    ResourceSet::getValue<int>(soci::session &sql, const std::string_view classname,
                                const std::string_view parameter) {
        using namespace soci;
        int a;
        sql << "SELECT value FROM " << intTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(a);
        if (sql.got_data())
            return a;

        return std::nullopt;
    }

    template<>
    inline std::optional<FontSelection>
    ResourceSet::getValue<FontSelection>(soci::session &sql, const std::string_view classname,
                                         const std::string_view parameter) {
        using namespace soci;
        int a;
        sql << "SELECT value FROM " << intTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(a);
        if (sql.got_data()) {
            if (a >= 0 && a < static_cast<int>(FontSelection::COUNT)) {
                return static_cast<FontSelection>(a);
            }
        }
        return std::nullopt;
    }

    template<>
    inline std::optional<FontSize>
    ResourceSet::getValue<FontSize>(soci::session &sql, const std::string_view classname,
                                    const std::string_view parameter) {
        using namespace soci;
        int a;
        sql << "SELECT value FROM " << intTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(a);
        if (sql.got_data()) {
            if (a >= 0 && a < static_cast<int>(FontSize::COUNT)) {
                return static_cast<FontSize>(a);
            }
        }
        return std::nullopt;
    }

    template<>
    inline std::optional<std::array<int,2>>
    ResourceSet::getValue(soci::session &sql, const std::string_view classname, const std::string_view parameter) {
        using namespace soci;
        row r;
        sql << "SELECT a,b FROM " << intPairTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(r);
        if (sql.got_data() && r.get_indicator(0) && r.get_indicator(1)) {
            std::array<int,2> result{};
            result[0] = r.get<int>(0);
            result[1] = r.get<int>(1);
            return result;
        }
        return std::nullopt;
    }

//    template<>
//    inline std::optional<ValuePair<int>>
//    ResourceSet::getValue<ValuePair<int>>(soci::session &sql, const std::string_view classname,
//                                          const std::string_view parameter) {
//        using namespace soci;
//        row r;
//        sql << "SELECT a,b FROM " << intPairTable << " WHERE class = \"" << classname << "\" AND parameter = \""
//            << parameter << "\" LIMIT 1", into(r);
//        if (sql.got_data() && r.get_indicator(0) && r.get_indicator(1))
//            return ValuePair<int>{r.get<int>(0), r.get<int>(1)};
//        return std::nullopt;
//    }

    template<>
    inline std::optional<std::array<int,4>>
    ResourceSet::getValue(soci::session &sql, const std::string_view classname, const std::string_view parameter) {
        using namespace soci;
        row r;
        sql << "SELECT a,b,c,d FROM " << intQuadTable << " WHERE class = \"" << classname << "\" AND parameter = \""
            << parameter << "\" LIMIT 1", into(r);
        if (sql.got_data()) {
            std::array<int,4> result{};
            result[0] = r.get<int>(0);
            result[1] = r.get<int>(1);
            result[2] = r.get<int>(2);
            result[3] = r.get<int>(3);
            return result;
        }
        return std::nullopt;
    }
//
//    template<>
//    inline std::optional<ValueQuad<int>>
//    ResourceSet::getValue<ValueQuad<int>>(soci::session &sql, const std::string_view classname,
//                                          const std::string_view parameter) {
//        using namespace soci;
//        int a, b, c, d;
//        row r;
//        sql << "SELECT a,b,c,d FROM " << intQuadTable << " WHERE class = \"" << classname << "\" AND parameter = \""
//            << parameter << "\" LIMIT 1", into(r);
//        if (sql.got_data())
//            return ValueQuad<int>{r.get<int>(0), r.get<int>(1), r.get<int>(2), r.get<int>(3)};
//        return std::nullopt;
//    }

    template<>
    inline std::optional<SDL_Color>
    ResourceSet::getValue<SDL_Color>(soci::session &sql, const std::string_view classname,
                                     const std::string_view parameter) {
        if (auto color = getValue<std::array<int,4>>(sql, classname, parameter); color.has_value()) {
            return SDL_Color{static_cast<Uint8>(color.value()[0]), static_cast<Uint8>(color.value()[1]),
                             static_cast<Uint8>(color.value()[2]), static_cast<Uint8>(color.value()[3])};
        }
        return std::nullopt;
    }

    template<>
    inline std::optional<Margin<int>>
    ResourceSet::getValue<Margin<int>>(soci::session &sql, const std::string_view classname, const std::string_view parameter) {
        if (auto margin = getValue<std::array<int,4>>(sql, classname, parameter); margin.has_value()) {
            return Margin<int>(margin.value());
        }
        return std::nullopt;
    }

    template<>
    inline std::optional<Padding<int>>
    ResourceSet::getValue<Padding<int>>(soci::session &sql, const std::string_view classname, const std::string_view parameter) {
        if (auto padding = getValue<std::array<int,4>>(sql, classname, parameter); padding.has_value()) {
            return Padding<int>(padding.value()[0], padding.value()[1], padding.value()[2], padding.value()[3]);
        }
        return std::nullopt;
    }

    class RoseLibRes : public ResourceSet {
    protected:
        RoseLibRes() = default;

    public:
        static constexpr Resources::database_signature signature
                {{
                         "Resource",
                         "Library",
                         "Rose"
                 }};

        static RoseLibRes &get() {
            static RoseLibRes roseLibRes{};
            return roseLibRes;
        }
    };

}

