/**
 * @file SDL++.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-13
 */

#pragma once

#include <array>
#include <chrono>
#include <filesystem>
#include <thread>
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_ttf.h>
#include "Utilities.h"
#include "ObjectGuard.h"

namespace rose {
    void SetDefaultSignalHandlers();

    enum class Orientation {
        Horizontal,
        Vertical,
    };

    /**
     * @brief A string_value is something that can be used to construct a std::string.
     * @tparam S The type to test.
     */
    template<class S>
    concept string_value = std::is_constructible_v<std::string, S>;

    namespace fs = ::std::filesystem;

    /**
     * @brief A file_path_value is either a string, which will be converted to a path, or an actual
     * std::filesystem::path.
     * @tparam S The type to test.
     */
    template<class S>
    concept file_path_value = string_value<S> || std::is_same_v<S, fs::path>;

    /**
     * @brief Type trait for determining if the shared_ptr<type> Derived is a shared pointer to a type derived
     * from Base. This evaluates to false.
     * @tparam Base The tested Base class.
     * @tparam Derived The tested shared_ptr<type> where type the test is std::is_base_of<Base,shared_ptr<type>>
     */
    template<class Base, class Derived>
    struct is_shared_ptr_base_of {
        static constexpr bool value{std::false_type::value};
    };

    /**
     * @brief Type trait to determine if in a shared_ptr<Derived>, Derived is a derived type of Base.
     * @details This evaluates to std::is_base_of_v<Base,Derived>
     * @tparam Base The tested base class
     * @tparam Derived The tested derived class.
     */
    template<class Base, class Derived>
    struct is_shared_ptr_base_of<Base,std::shared_ptr<Derived>> {
        inline static constexpr bool value{ std::is_base_of_v<Base,Derived>};
    };

    /**
     * @brief Type trait to determine if in a shared_ptr<Derived>, Derived is a derived type of Base.
     * @details This evaluates to std::is_base_of_v<Base,Derived>
     * @tparam Base The tested base class
     * @tparam Derived The tested derived class.
     */
    template<class Base, class Derived>
    inline static constexpr bool is_shared_ptr_base_of_v = is_shared_ptr_base_of<Base,Derived>::value;

    /**
     * @brief A concept of an object being a base class of the type of a std::shared_ptr.
     * @tparam Base The tested base class
     * @tparam Derived The tested derived class.
     */
    template<class Base, class Derived>
    concept pointer_base_of = is_shared_ptr_base_of_v<Base,Derived>;

    class ErrorStatus {
    public:
        enum Status {
            OK, SDL_ERROR, KEY_NOT_FOUND, EMPTY_UNIQUE_POINTER, NOT_ACCESS_TARGET, DOMAIN_ERROR, GFX_ERROR, COUNT
        };

        static constexpr std::array<std::string_view,Status::COUNT> ErrorStrings{
                    "Ok",
                    "SDL Library Error",
                    "Key not found in container",
                    "Empty std::unique_ptr",
                    "Texture is not a render access target.",
                    "Domain error, value out of range or wrong type.",
                    "SDL2_gfx libarary error.",
        };

    protected:
        bool mThrow{true};
        bool mLog{};
        Status mStatus{OK};
        std::string_view mFile{};
        std::string sdlErrorString{};
        int mLine{0};

        ErrorStatus() = default;

    public:
        static ErrorStatus& get() {
            static ErrorStatus errorStatus{};
            return errorStatus;
        }

        [[nodiscard]] explicit operator bool () const noexcept {
            return mStatus == OK;
        }

        bool setStatus(Status status, const char *file, int line) {
            mStatus = status;
            if (status == SDL_ERROR)
                sdlErrorString = SDL_GetError();
            else
                sdlErrorString.clear();
            mFile = file;
            mLine = line;

            if (mThrow && mStatus != OK) {
                throw std::runtime_error(rose::StringCompositor(mFile, ':', mLine, " -- ", ErrorStrings[mStatus], sdlErrorString));
            }
            return false;
        }

        void throwException() {
            throw std::runtime_error(rose::StringCompositor(mFile, ':', mLine, " -- ", ErrorStrings[mStatus], sdlErrorString));
        }

        void clear() {
            mStatus = OK;
        }
    };

#define RoseError(status) ErrorStatus::get().setStatus((status), __FILE__, __LINE__)

    class SdlCursorDestroy {
    public:
        /**
         * @brief Call the SDL API to destroy an SDL_Cursor.
         * @param sdlCursor A pointer to the SDL_Window to destroy.
         */
        void operator()(SDL_Cursor *sdlCursor) {
            std::cerr << "SdlWindowDestroy\n";
            if (sdlCursor) {
                SDL_FreeCursor(sdlCursor);
            }
        }
    };

    class SdlCursor : public std::unique_ptr<SDL_Cursor, SdlCursorDestroy> {
    public:
        SdlCursor() = default;
        ~SdlCursor() = default;

        explicit SdlCursor(SDL_Cursor* cursor) : std::unique_ptr<SDL_Cursor, SdlCursorDestroy>(cursor) {}

        explicit SdlCursor(SDL_SystemCursor id) : SdlCursor() {
            auto cursor = SDL_CreateSystemCursor(id);
            if (cursor)
                reset(cursor);
            else
                RoseError(ErrorStatus::SDL_ERROR);
        }

        void set() {
            SDL_SetCursor(get());
        }
    };

}
