/**
 * @file SdlWindow.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-20
 */

#pragma once

#include <memory>
#include <SDL.h>
#include "Metrics.h"
#include "SDL++.h"

namespace rose::gm {

    /**
     * @class SdlWindow
     * @brief A C++ interface to SDL_Window
     */
    class SdlWindow {

    protected:
        /**
         * @brief A functor to destroy an SDL_Window in a std::unique_ptr (rose::sdl::Window).
         */
        class SdlWindowDestroy {
        public:
            /**
             * @brief Call the SDL API to destroy an SDL_Window.
             * @param sdlWindow A pointer to the SDL_Window to destroy.
             */
            void operator()(SDL_Window *sdlWindow) {
                std::cerr << "SdlWindowDestroy\n";
                if (sdlWindow) {
                    SDL_DestroyWindow(sdlWindow);
                }
            }
        };

        using SdlWindowPtr [[maybe_unused]] = std::unique_ptr<SDL_Window, SdlWindowDestroy>;   //!< An SDL_Window unique pointer

        SdlWindowPtr mWindow{};

        bool mStatus{true};

    public:

        SdlWindow() = default;

        SdlWindow(const SdlWindow &context) = delete;

        SdlWindow(SdlWindow &&context) noexcept = default;

        SdlWindow &operator=(const SdlWindow &context) = delete;

        SdlWindow &operator=(SdlWindow &&context) = default;

        template<class S>
        SdlWindow(S title, const Size<int> &size, const Position<int> &position = PositionCentered,
                  uint32_t flags = 0)
                : SdlWindow() {
            static_assert(std::is_constructible_v<std::string, S>,
                          "Must be able to construct a std::string from title.");
            std::string sTitle{title};
            mWindow.reset(SDL_CreateWindow(sTitle.c_str(), position.x, position.y, size.w, size.h, flags));
            if (!mWindow)
                mStatus = RoseError(ErrorStatus::SDL_ERROR);
        }

        [[nodiscard]] explicit operator bool() const noexcept {
            return mStatus && mWindow;
        }

        [[nodiscard]] SDL_Window *get() const noexcept {
            return mWindow.get();
        }
    };
}

