/**
 * @file Position.b
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-27
 */

#pragma once

#include <compare>
#include <concepts>
#include <type_traits>
#include "SDL++.h"

namespace rose {
    /**
 * @struct Size
 * @brief A size in integer dimensions.
 * @details The width (w) is in the direction of the x axis, the height (h) is in the direction of the y axis.
 */
    template<typename T=int>
    requires std::integral<T> || std::floating_point<T>
    struct Size {
        T w{}, h{};
        constexpr Size() noexcept = default;
        constexpr Size(const Size &p) noexcept = default;
        constexpr Size(Size &&p)  noexcept = default;
        constexpr Size& operator=(const Size &p) noexcept = default;
        constexpr Size& operator=(Size &&p)  noexcept = default;
        constexpr explicit Size(T p) noexcept : w(p), h(p) {}

        template<typename O>
        constexpr Size(O W, O H) noexcept : w(W), h(H) {}

        [[nodiscard]] constexpr Size<T> operator + (const Size<T>& other) const noexcept {
            return Size<T>(this->w+other.w, this->h+other.h);
        }

        [[nodiscard]] constexpr Size<T> operator - (const Size<T>& other) const noexcept {
            return Size<T>(this->w-other.w, this->h-other.h);
        }

        [[nodiscard]] constexpr Size<T> operator / (const T& value) const {
            return Size<T>(this->w/value, this->h/value);
        }

        [[nodiscard]] constexpr bool operator < (const Size<T>& other) const noexcept {
            return this->w < other.w && this->h < other.h;
        }

        /**
         * @brief Three way comparison of two Positions.
         * @details If corresponding coordinates of this and other are equivalent, then this is equivalent to other.
         * If one co-ordinate of this is less than its correspondent of other and the orthogonal
         * is less than or equal to its correspondent then this is less than other. If one co-ordinate of this is
         * greater than its correspondent of other and the orthogonal is greater than or equal to its
         * correspondent then this is greater than other. Otherwise the this and other are unordered.
         * @param other The other Position
         * @return A std::partial_ordering value.
         */
        constexpr std::partial_ordering operator <=>(const Size& other) const noexcept {
            auto wo = w <=> other.w;
            auto ho = h <=> other.h;

            if ((wo < 0 && ho <= 0) || (wo <= 0 && ho < 0))
                return std::partial_ordering::less;
            if ((wo > 0 && ho >= 0) || (wo >= 0 && ho > 0))
                return std::partial_ordering::greater;
            if (wo == 0 && ho == 0)
                return std::partial_ordering::equivalent;

            return std::partial_ordering::unordered;
        }

        constexpr bool operator == (const Size& other) const noexcept {
            return operator<=>(other) == 0;
        }

        constexpr bool operator != (const Size& other) const noexcept {
            return operator<=>(other) != 0;
        }

        [[nodiscard]] T& primary(Orientation orientation) {
            return orientation == Orientation::Horizontal ? w : h;
        }

        [[nodiscard]] T& secondary(Orientation orientation) {
            return orientation == Orientation::Vertical ? w : h;
        }

        [[nodiscard]] const T& primary(Orientation orientation) const {
            return orientation == Orientation::Horizontal ? w : h;
        }

        [[nodiscard]] const T& secondary(Orientation orientation) const {
            return orientation == Orientation::Vertical ? w : h;
        }

        std::ostream& printOn(std::ostream& strm) const {
            return strm << '[' << w << ',' << h << ']';
        }
    };

    static Size<int> SizeZero;
}
