/**
 * @file SystemData.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-05
 */

#pragma once

#include "Sdl++/Signals.h"
#include "TimerTick.h"

namespace rose {

    using CPU_LoadProtocol = Protocol<float,float>;     ///< Signal Protocol for CPU Load Average and Process CPU Use.
    using CPU_TemperatureProtocol = Protocol<float>;    ///< Signal Protocol for CPU Temperature

    enum CpuTimeIndex : std::size_t {
        USER = 0, NICE, SYSTEM, IDLE, IOWAIT, IRQ, SOFTIRQ, STEAL, GUEST, GUEST_NICE, ARRAY_SIZE
    };

    /**
     * @class SystemData
     * @brief A singleton object that gathers CPU Temperature, Load Average and Process CPU use for transmission
     * on a Signals::Protocol.
     */
    class SystemData {
    protected:
        bool mHasTemperatureDevice{true};       ///< True if temperature data is found.

        /**
         * @brief Read the CPU temperature from the system.
         */
        void readCPUTemperature();

        /**
         * @brief Read processor usage and estimate this process' contribution.
         */
        void readProcessUsage();

        /**
         * @brief Count the number of CPUs in the system.
         */
        void cpuCount();

        std::array<uint64_t,CpuTimeIndex::ARRAY_SIZE> mCpuTime{};       ///< CPU current times
        std::array<uint64_t,CpuTimeIndex::ARRAY_SIZE> mPastCpuTime{};   ///< CPU past times

        int mCpuCount{0};            ///< The number of CPUs in the system.
        float mCpuTimeUse{};         ///< The current CPU time use
        float mCpuTotalTime{};       ///< The total CPU time available.
        int mProcTimeUse{};          ///< The current process time use
        int mProcTimeStart{};        ///< The process time use at interval start

        float mUsage{0.f};           ///< The process usage as percent of a CPU.
        float mTemperature{0.f};     ///< The cpu temperature.

        static constexpr std::string_view ProcSelfStat = "/proc/self/stat";     ///< The process stats path
        static constexpr std::string_view ProcStat = "/proc/stat";              ///< The system stats apth
        static constexpr std::string_view ProcCpuInfo = "/proc/cpuinfo";        ///< The CPU info path

        TickProtocol::slot_type second_slot{};

        SystemData();

    public:
        CPU_LoadProtocol::signal_type cpuLoad{};                ///< The CPU Load signal.

        CPU_TemperatureProtocol::signal_type cpuTemperature{};  ///< The CPU Temperature signal.

        ~SystemData() = default;

        /**
         * @brief Get a reference to the SystemData singleton.
         * @return SystemData reference.
         */
        static SystemData& get() {
            static SystemData systemData{};
            return systemData;
        }

        using TemperatureSignal = std::array<float,3>;  ///< The type used for system temperature signal.
        Signal<TemperatureSignal> txTemperature{};      ///< The Signal object to transmit temperature.

        using ProcessSignal = std::array<float,3>;      ///< The type used for process usage signal.
        Signal<ProcessSignal> txProcess{};              ///< The Signal object to transmit process usage.

        using SystemSignal = std::array<float,3>;       ///< The type used for the system usage signal.
        Signal<SystemSignal> txSystem{};
    };
}
