/**
 * @file Text.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-24
 */

#pragma once

#include <utility>

#include "Sdl++/Font.h"
#include "Theme.h"
#include "Widget.h"
#include "Sdl++/Metrics.h"
#include "Sdl++/SdlGraphicsModel.h"

namespace rose {

    struct FontPointSize {
        FontPointSize() = default;
        ~FontPointSize() = default;
        FontPointSize(const FontPointSize&) = default;
        FontPointSize(FontPointSize&&) = default;
        FontPointSize& operator = (const FontPointSize&) = default;
        FontPointSize& operator = (FontPointSize&&) = default;
        using point_size_type = int;
        point_size_type pt{};
    };

    /**
     * @class TypeBlock
     * @brief A unit of type.
     */
    class TypeBlock {
    public:
        friend class LineTypeSetter;
        friend class ColumnTypeSetter;

        enum Align {
            SUPERSCRIPT,
            TOP,
            BASELINE,
            SUBSCRIPT,
            LEFT,
            CENTRE,
            RIGHT,
        };

        enum Kerning {
            NONE,
            QUARTER_EN,
            HALF_EN,
            EN,
            EM,
        };

        template<class Parameter>
        void setParameter(Parameter) {
            std::cerr << '[' << __FILE__ << ':' << __LINE__ << "] Parameter not supported\n";
        }

        template<class Parameter>
        requires string_value<Parameter>
        void setParameter(Parameter parameter) {
            mText = parameter;
        }

    protected:
        Theme& mTheme;
        gm::Texture mTexture{};
        Align mAlign{BASELINE};
        Size<int> mSize{};
        Position<int> mPosition{};
        FontPointSize mFontPointSize{};
        FontSelection mFontSelection{FontSelection::REGULAR};
        FontPointer mFont{};
        FontMetrics mFontMetrics{};
        Kerning mKerning{NONE};
        int eN{}, eM{};
        SDL_Color mInk{color::RGBA::OpaqueBlack.toSdlColor()};
        std::string mText{};
        uint32_t mCodePoint{};

    public:
        TypeBlock() : mTheme(Theme::get()) {}

        explicit TypeBlock(const SDL_Color& ink) : mTheme(Theme::get()), mInk(ink) {}

        virtual ~TypeBlock() = default;

        virtual void layout(gm::Context &context);

        virtual void draw(gm::Context &context, const Position<int>& lineOrigin);
    };

    template<>
    inline void TypeBlock::setParameter<FontSize>(FontSize parameter) {
        mFontPointSize.pt = mTheme.fontSize[asSize(parameter)];
    }

    template<>
    inline void TypeBlock::setParameter<FontPointSize>(FontPointSize fontPointSize) {
        mFontPointSize = fontPointSize;
    }

    template<>
    inline void TypeBlock::setParameter<FontSelection>(FontSelection parameter) { mFontSelection = parameter; }

    template<>
    inline void TypeBlock::setParameter<TypeBlock::Kerning>(TypeBlock::Kerning parameter) { mKerning = parameter; }

    template<>
    inline void TypeBlock::setParameter<SDL_Color>(SDL_Color parameter) { mInk = parameter; }

    template<>
    inline void TypeBlock::setParameter<TypeBlock::Align>(TypeBlock::Align parameter) { mAlign = parameter; }

    template<>
    inline void TypeBlock::setParameter<uint32_t>(uint32_t parameter) { mCodePoint = parameter; }

    class ColumnTypeSetter;

    /**
     * @class LineTypeSetter
     * @brief Manages a number of TypeBlocks in a line to create an impression texture for rendering.
     */
    class LineTypeSetter : public TypeBlock {
    protected:
        std::vector<std::shared_ptr<TypeBlock>> mContent{};

    public:
        LineTypeSetter() = default;

        ~LineTypeSetter() override = default;

        auto at(std::vector<std::shared_ptr<TypeBlock>>::size_type n) {
            return mContent.at(n);
        }

        auto content() { return mContent; }

        auto size() { return mSize; }

        template<typename...Args>
        void emplace(Args...args) {
            auto& typeBlock = mContent.emplace_back(std::make_shared<TypeBlock>());
            (typeBlock->template setParameter(args),...);
        }

        template<typename...Args>
        std::shared_ptr<ColumnTypeSetter> emplaceColumn(Args...args) {
            auto& column = mContent.emplace_back(std::make_shared<ColumnTypeSetter>());
            (column->template setParameter(args), ...);
            return std::dynamic_pointer_cast<ColumnTypeSetter>(column);
        }

        void layout(gm::Context &context) override;

        void draw(gm::Context &context, const Position<int>& lineOrigin) override {
            for (auto & block : mContent) {
                block->draw(context, lineOrigin + mPosition);
            }
        }
    };

    /**
     * @class ColumnTypeSetter
     * @brief Manages a number of TypeBlocks in a column to create an impression texture for rendering.
     */
    class ColumnTypeSetter : public TypeBlock {
    protected:
        std::vector<std::shared_ptr<TypeBlock>> mContent{};

    public:
        ColumnTypeSetter() = default;

        ~ColumnTypeSetter() override = default;

        auto at(std::vector<std::shared_ptr<TypeBlock>>::size_type n) {
            return mContent.at(n);
        }

        auto content() { return mContent; }

        template<typename...Args>
        void emplace(Args...args) {
            auto& typeBlock = mContent.emplace_back(std::make_shared<TypeBlock>());
            ((typeBlock->template setParameter(args)), ...);
        }

        void layout(gm::Context &context) override;

        void draw(gm::Context &context, const Position<int>& lineOrigin) override {
            for (auto & block : mContent) {
                block->draw(context, lineOrigin + mPosition);
            }
        }

        template<typename...Args>
        std::shared_ptr<LineTypeSetter> emplaceLine(Args...args) {
            auto& column = mContent.emplace_back(std::make_shared<LineTypeSetter>());
            ((column->template setParameter(args)), ...);
            return std::dynamic_pointer_cast<LineTypeSetter>(column);
        }
    };

    /**
     * @class Text
     * @brief
     */
    class Text : public Widget {

    protected:
        static constexpr std::string_view TextClass{"Text"};
        FontPointer mFont{};
        FontSelection mFontSelection{FontSelection::REGULAR};
        FontPointSize mFontPointSize{};
        gm::Texture mTexture{};

        std::string mText{""};

    private:
        void getResources();

    public:
        Text() : Widget(property::Class, TextClass) {
            getResources();
        }

        Text(const property_key key, const property_value &value) : Widget(key, value) {
            getResources();
        }

        ~Text() override = default;

        /**
         * @brief Type directed parameter setting.
         * @details Derived classes must implement this function to take parameter types that they handle, then
         * call this function on their base class for those parameter types they don't handle.
         * @tparam Parameter The type of the parameter.
         * @param parameter The value of the parameter.
         */
        template<class Parameter>
        void setParameter(Parameter parameter) {
            Widget::setParameter(parameter);
        }

        template<class Parameter>
        requires string_value<Parameter>
        void setParameter(Parameter parameter) {
            mText = parameter; mTexture.reset(nullptr);
        }

        void setFontSelection(FontSelection) {  }

        void
        draw(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize) override;

        void
        layout(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize) override;

    };

    template<>
    inline void Text::setParameter<FontPointer>(FontPointer parameter) {
        mFont = std::move(parameter);
        mTexture.reset(nullptr);
    }

    template<>
    inline void Text::setParameter<FontSelection>(FontSelection parameter) {
        mFontSelection = parameter;
        mFont = nullptr;
    }

    template<>
    inline void Text::setParameter<FontPointSize>(FontPointSize parameter) {
        mFontPointSize.pt = parameter.pt;
    }

    template<>
    inline void Text::setParameter<FontSize>(FontSize parameter) {
        mFontPointSize.pt = mTheme.fontSize[asSize(parameter)];
        mFont = nullptr;
    }

    using TextPtr = std::shared_ptr<Text>;
}

