/**
 * @file TextureCache.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-18
 */

#pragma once

#include <map>
#include "Sdl++/SDL++.h"
#include "Rose.h"
#include "Sdl++/Metrics.h"
#include "Sdl++/SdlGraphicsModel.h"
#include "Sdl++/Texture.h"
#include "Sdl++/WebCache.h"

namespace rose {

    namespace fs = std::filesystem;

    /**
     * @class TextureCache
     * @brief Cache Texture objects.
     * @details Each Texture is associated with a unique identifier which should be taken from rose::UniqueId::userid(),
     * a name which can be used to annotate displays or search for the Texture but is not required to be unique, and a
     * texture class number which is used to group Textures by a user defined category system. The category numbers
     * should also be taken from rose::UniqueId::userid() which will never return 0 so 0 can be used to mean
     * uncategorized.
     */
    class TextureCache {
    protected:
        struct CachedTexture {
            unique_id_type id;
            std::string name;
            unique_id_type textureClass{0};
            gm::Texture texture{};
            gm::Surface surface{};

            template<typename S>
            requires string_value<S>
            CachedTexture(unique_id_type id, S name, unique_id_type textureClass)
                : id(id), name(name), textureClass(textureClass) {}

            template<typename S>
            requires string_value<S>
            CachedTexture(unique_id_type id, S name, unique_id_type textureClass, gm::Texture t)
                    : id(id), name(name), textureClass(textureClass), texture(std::move(t)) {}

            template<typename S>
            requires string_value<S>
            CachedTexture(unique_id_type id, S name, unique_id_type textureClass, gm::Surface s)
                    : id(id), name(name), textureClass(textureClass), surface(std::move(s)) {}
        };

        using TextureCachePair = std::pair<const unique_id_type,CachedTexture>;
        using TextureCacheMap = std::map<unique_id_type,CachedTexture>;

        /// A directory that may be used to associate names with image class ids
        std::map<std::string,unique_id_type> mCacheDirectory{};

        WebCacheProtocol::slot_type mWebCacheSlot{};

        WebCache &mWebCache;
        std::vector<unique_id_type> mWebCacheSubscriptions{};

        TextureCacheMap mTextures{};

        TextureCache() : mWebCache(WebCache::get()) {}

        static bool updateFromSurface(gm::Context& context, TextureCacheMap::iterator iterator);

        static bool resolveTexture(gm::Context& context, TextureCacheMap::iterator iterator);

    public:
        /**
         * @brief Get the singleton instance of the TextureCaceh.
         * @return a TextureCache&
         */
        static TextureCache& get() {
            static TextureCache textureCache{};
            return textureCache;
        }

        /**
         * @brief Determine if a particular id is cached.
         * @param id The id to look for.
         * @return true if the ide is found.
         */
        [[nodiscard]] bool cached(unique_id_type id) const {
            return mTextures.find(id) != mTextures.end();
        }

        /**
         * @brief Emplace a new texture entry into the map, unless the key already exists then update the entry.
         * @details There are two version of this function, one emplaces a Texture, the other a Surface.
         * @param key The key, should be a unique_id_type from rose::UniqueId::userid().
         * @param name A name that can be used to label the image.
         * @param texture The image texture. The texture is moved leaving the original empty.
         * @param textureClass A set of flags to allow searching for groups of images.
         * @return true if emplaced, false on error. Check Rose::ErrorStatus for details.
         */
        template<class S>
        requires string_value<S>
        void emplaceOrUpdate(unique_id_type key, S name, gm::Texture texture, unique_id_type textureClass = 0) {
            auto [it, res] = mTextures.try_emplace(key, key, name, textureClass);
            if (res) {
                it->second.texture = std::move(texture);
            } else {
                it->second.name = std::move(name);
                it->second.textureClass = textureClass;
                it->second.texture = std::move(texture);
                it->second.surface.reset(nullptr);
            }
        }

        template<class S>
        requires string_value<S>
        void emplaceOrUpdate(unique_id_type key, S name, gm::Surface surface, unique_id_type textureClass = 0) {
            auto [it, res] = mTextures.try_emplace(key, key, name, textureClass);
            if (res) {
                it->second.surface = std::move(surface);
            } else {
                it->second.name = std::move(name);
                it->second.textureClass = textureClass;
                it->second.surface.reset(nullptr);
                it->second.surface = std::move(surface);
            }
        }

        template<typename S>
        requires string_value<S>
        bool updateOrEmplace(unique_id_type key, S name = "", unique_id_type textureClass = 0) {
            gm::Texture texture{};
            auto [it, res] = mTextures.try_emplace(key, key, name, textureClass);
            if (res) {
                std::string_view nameView{name};
                if (!nameView.empty())
                    it->second.name = nameView;
                if (textureClass)
                    it->second.textureClass = textureClass;
            }
            return true;
        }

        /**
         * @brief Update the cache entry with a Surface that will be converted to a Texture on the next call that
         * operates on the entry with a gm::Context.
         * @details This clears the texture entry but leaves all other data untouched.
         * @param surface The surface to update.
         * @param name If name is not an empty string, and the record name is an empty string, the record name will
         * be st to the name argument.
         * @return true on success, false on failure.
         */
        template<typename S>
        requires string_value<S>
        bool updateSurface(unique_id_type key, gm::Surface surface, S name = "") {
            if (mTextures.find(key) != mTextures.end()) {
                if (auto it = mTextures.find(key); it != mTextures.end()) {
                    it->second.surface = std::move(surface);
                    it->second.texture.reset(nullptr);
                    if (it->second.name.empty() && !name.empty())
                        it->second.name = std::move(name);
                    return true;
                }
            }
            return false;
        }

        /**
         * @brief Load a texture from the WebCache.
         * @param key Identifies which request the Texture will be created from and the id of the Texture in
         * this cache.
         * @return true if the image was loaded. Check Rose::ErrorStatus for details.
         */
        bool subscribeToWebCache(unique_id_type key, unique_id_type classType = 0) {
            if (!mWebCacheSlot) {
                mWebCacheSlot = WebCacheProtocol::createSlot();
                mWebCacheSlot->receiver = [&](unique_id_type id, long status) {
                    if (status == 200 || (status == 302 && !cached(id))) {
                        if (std::find(mWebCacheSubscriptions.begin(), mWebCacheSubscriptions.end(), id) != mWebCacheSubscriptions.end()) {
                            if (auto filePath = mWebCache.requestFilePath(id); filePath) {
                                gm::Surface surface;
                                surface.loadFile(filePath.value());
                                if (surface) {
                                    auto name = filePath->filename().string();
                                    updateSurface(id, std::move(surface), name);
                                }
                            }
                        }
                    }
                };
            }

            if (std::find(mWebCacheSubscriptions.begin(), mWebCacheSubscriptions.end(), key) == mWebCacheSubscriptions.end()) {
                updateOrEmplace(key, "", classType);
                mWebCacheSubscriptions.push_back(key);
                if (auto filePath = mWebCache.requestFilePath(key); filePath) {
                    gm::Surface surface;
                    surface.loadFile(filePath.value());
                    if (surface) {
                        auto name = filePath->stem().string();
                        std::cout << name << '\n';
                        updateSurface(key, std::move(surface), name);
                    }
                }
                return true;
            }

            return false;
        }

        /**
         * @brief Update the Texture in the cache if the key exits.
         * @param key The key of the texture to update.
         * @param texture The replacement texture. The texture is moved leaving the original empty.
         * @return true if the texture was replaced, false if the key does not exist.
         */
        bool update(unique_id_type key, gm::Texture texture) {
            if (!texture)
                return RoseError(ErrorStatus::EMPTY_UNIQUE_POINTER);

            if (auto it = mTextures.find(key); it != mTextures.end()) {
                it->second.texture = std::move(texture);
                return true;
            }

            return RoseError(ErrorStatus::KEY_NOT_FOUND);
        }

        /**
         * @brief Render a cached Texture using a provided context.
         * @details Copies the full texture mapping into the full destination.
         * @param key The key to find the Texture.
         * @param context The graphics context.
         * @return true on success, false on failure. See Rose::ErrorStatus for details.
         */
        bool draw(unique_id_type key, gm::Context& context);

        /**
         * @brief Render a cached Texture using a provided context.
         * @details Copies the full texture mapping into the destination Rectangle.
         * @param key The key to find the Texture.
         * @param context The graphics context.
         * @param dst The destination rectangle.
         * @return 0 when OK, negative on an SDL Library error, 1 if there is no key value in the cache.
         */
        bool draw(unique_id_type key, gm::Context& context, const Rectangle<int>& dst);

        /**
         * @brief Render a cached Texture using a provided context.
         * @details Copies the source rectangle from the Texture to the destination rectangle on the target.
         * @param key The key to find the Texture.
         * @param context The graphics context.
         * @param src The source rectangle.
         * @param dst The destination rectangle.
         * @return 0 when OK, negative on an SDL Library error, 1 if there is no key value in the cache.
         */
        bool draw(unique_id_type key, gm::Context& context, const Rectangle<int>& src, const Rectangle<int>& dst) {
            if (auto it = mTextures.find(key); it != mTextures.end()) {
                if (context.renderCopy(it->second.texture, src, dst))
                    return RoseError(ErrorStatus::SDL_ERROR);
                return true;
            }
            return RoseError(ErrorStatus::KEY_NOT_FOUND);
        }

        /**
         * @brief Get the Size of an image given the image id.
         * @return A std::optional with the Size, or empty if there is no image associated with the id.
         */
        [[nodiscard]] std::optional<Size<int>> size(unique_id_type key) const {
            if (auto it = mTextures.find(key); it != mTextures.end()) {
                if (it->second.texture)
                    return it->second.texture.getSize();
                else if (it->second.surface)
                    return Size<int>(it->second.surface->w, it->second.surface->h);
            }
            return std::nullopt;
        }

        /**
         * @brief Gind the image id give the image name.
         * @param name The requested name.
         * @return The associated image id or 0
         */
        [[nodiscard]] unique_id_type findByName(const std::string& name) const {
            if (auto it = std::find_if(mTextures.begin(), mTextures.end(), [&name](
                        const TextureCachePair& pair) -> bool{
                return pair.second.name == name;
            }); it != mTextures.end()) {
                return it->first;
            }
            return 0;
        }

        /**
         * @brief Get a list of cached images that have been assigned to a particular image class.
         * @param id The unique_id of the class type.
         * @return A vector of image unique_ids of the images belongin to the class.
         */
        [[nodiscard]] std::vector<unique_id_type> listImageClass(unique_id_type id) const {
            std::vector<unique_id_type> result{};

            if (id == 0)
                return result;

            std::for_each(mTextures.begin(), mTextures.end(), [&result,id](
                    const TextureCachePair& pair){
                if (pair.second.textureClass == id)
                    result.push_back(pair.first);
            });
            return result;
        }

        [[nodiscard]] std::vector<unique_id_type> listImageClass(const std::vector<unique_id_type>& list) const {
            std::vector<unique_id_type> result{};

            for (auto &imgClass : list) {
                if (auto add = listImageClass(imgClass); !add.empty())
                    result.insert(result.end(), std::make_move_iterator(add.begin()), std::make_move_iterator(add.end()));
            }

            return result;
        }

        template<class S>
        requires string_value<S>
        [[nodiscard]] unique_id_type directory(S name) {
            if (auto entry = std::find_if(mCacheDirectory.begin(), mCacheDirectory.end(), [&name](
                    const std::pair<const std::string&, unique_id_type>& pair){
                return pair.first == name;
            }); entry != mCacheDirectory.end()) {
                return entry->second;
            }

            auto newId = UniqueId::userId();
            mCacheDirectory.template emplace(name, newId);
            return newId;
        }

        /**
         * @brief Get a const reference to a Texture in the cache.
         * @param context The rendering context to convert a new surface if one has been emplaced.
         * @param key The Texture id.
         * @return A const reference to the cached Texture, or an empty Texture if key is not valid.
         */
        const gm::Texture & textureReference(gm::Context &context, unique_id_type key);
    };
}

