/**
 * @file Theme.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-24
 */

#pragma once

#include <string>
#include <string_view>
#include "Color.h"

namespace rose {

    template<class Value>
    inline constexpr std::size_t asSize(Value value) { return static_cast<std::size_t>(value); }

    enum class TextRenderStyle {
        Blended,        ///< Render text blended on a transparent background.
        Shaded,         ///< Render text shaded on a solid background.
        Solid           ///< Render text without blending or shading. Fastest but lowest quality.
    };

    enum class ColorSetIndex {
        INACTIVE,
        ACTIVE,
        DISABLED,
    };

    static constexpr std::string_view FontSelectionName = "FontSelection";
    enum class FontSelection : std::size_t {
        REGULAR,
        BOLD,
        ITALIC,
        BOLD_ITALIC,
        FIXED,
        FIXED_BOLD,
        FIXED_ITALIC,
        FIXED_BOLD_ITALIC,
        ENTYPO,
        COUNT
    };

    static constexpr std::string_view FontPointSizeName = "FontPointSize";
    static constexpr std::string_view FontSizeName = "FontSize";
    enum class FontSize : std::size_t {
        VERY_SMALL,
        SMALL,
        NORMAL,
        BUTTON,
        LARGE,
        VERY_LARGE,
        COUNT
    };

    enum class ColorIndex : std::size_t {
        BACKGROUND,
        FOREGROUND,
        TOP_BORDER,
        BOT_BORDER,
        LT_BORDER,
        RT_BORDER,
    };

    /**
     * @struct Theme
     * @brief Values used to determine the visual theme of the application.
     */
    class Theme {
    private:
        Theme() = default;

    public:

        using color_set = std::array<color::RGBA, 6>;
        using font_set = std::array<std::string_view, asSize(FontSelection::COUNT)>;
        using font_size = std::array<int, asSize(FontSize::COUNT)>;

        template<class Name>
        requires std::is_convertible_v<Name, std::string_view>
        constexpr Theme(TextRenderStyle textRenderStyle, Name name, font_set fontSet, font_size fontSize,
                        std::array<color_set, 3> colorSet) noexcept
                : textRenderStyle(textRenderStyle), name(name), fontNames(fontSet), fontSize(fontSize),
                  colorSet(colorSet) {}

        ~Theme() = default;

        static Theme &get() {
            static Theme theme{};
            return theme;
        }

        TextRenderStyle textRenderStyle{TextRenderStyle::Blended};
        std::string_view name{"default"};
        font_set fontNames{};
        font_size fontSize{};
        std::array<color_set, 3> colorSet{};

        explicit Theme(const Theme &) = delete;

        Theme(Theme &&) = delete;

        Theme &operator=(const Theme &) = default;

        Theme &operator=(Theme &&) = default;
    };

    static constexpr Theme DarkTheme{
            TextRenderStyle::Blended,
            "Dark",
            {
                    "FreeSans", "FreeSansBold", "FreeSansOblique", "FreeSansBoldOblique",
                    "FreeMono", "FreeMonoBold", "FreeMonoOblique", "FreeMonoBoldOblique",
                    "entypo"
            },
            {
                    10, 15, 20, 30, 40, 50
            },
            {{
                     { // Inactive
                             color::DarkBaseColor,
                             color::DarkTextColor,
                             color::DarkTopColor,
                             color::DarkBotColor,
                             color::DarkLeftColor,
                             color::DarkRightColor,
                     },
                     { // Active
                             color::DarkInvertColor,
                             color::DarkTextColor,
                             color::DarkBotColor,
                             color::DarkTopColor,
                             color::DarkRightColor,
                             color::DarkLeftColor
                     },
                     { // Disabled
                             color::DarkBaseColor,
                             color::DarkTextColor,
                             color::DarkBaseColor,
                             color::DarkBaseColor,
                             color::DarkBaseColor,
                             color::DarkBaseColor,
                     }
             }},
    };
}

