/**
 * @file Utilities.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-23
 */

#pragma once

#include <algorithm>
#include <iterator>
#include <sstream>
#include <vector>

#define FILE_LOC " -- ", __FILE__, ':', __LINE__    ///< Macro to create a location within a source file

namespace rose {
    /**
     * @class InputParser
     * @brief Parse command line arguments.
     */
    class [[maybe_unused]] InputParser {
    public:
        /**
         * @brief Constructor
         * @param argc The number of command line arguments passed to the application
         * @param argv The array of command line arguments.
         */
        [[maybe_unused]] InputParser(int &argc, char **argv) {
            for (int i = 1; i < argc; ++i)
                this->tokens.emplace_back(argv[i]);
        }

        /// @author iain
        [[maybe_unused]] [[nodiscard]] const std::string &getCmdOption(const std::string_view option) const {
            std::vector<std::string>::const_iterator itr;
            itr = std::find(this->tokens.begin(), this->tokens.end(), option);
            if (itr != this->tokens.end() && ++itr != this->tokens.end()) {
                return *itr;
            }
            static const std::string empty_string;
            return empty_string;
        }

        /// @author iain
        [[maybe_unused]] [[nodiscard]] bool cmdOptionExists(const std::string_view option) const {
            return std::find(this->tokens.begin(), this->tokens.end(), option)
                   != this->tokens.end();
        }

    private:
        std::vector<std::string> tokens;
    };

    /**
     * @class ReverseContainerView
     * @brief Provide a reverse view of a standard container type.
     * @tparam ContainerType The type of container
     */
    template<class ContainerType>
    class [[maybe_unused]] ReverseContainerView {
    protected:
        ContainerType &mContainer;      ///< A reference to the standard container.

    public:
        ReverseContainerView() = delete;

        /**
         * @brief Constructor.
         * @param container The standard container to provide the view of.
         * @param reverse If true reverse the container view, if false do not reverse.
         */
        explicit ReverseContainerView(ContainerType &container) : mContainer(container) {}

        /**
         * @brief Get the begin iterator for the reverse view
         * @return the standard container rbegin iterator
         */
        auto begin() {
            return std::rbegin(mContainer);
        }

        /**
         * @brief Get the end iterator for the reverse view
         * @return the standard container rend iterator
         */
        auto end() {
            return std::rend(mContainer);
        }
    };

    /**
     * @class ContainerView
     * @brief Provide a view of a standard container type.
     * @details Intended to be constructed by classes which have member containers.
     * @tparam ContainerType The type of container
     */
    template<class ContainerType>
    class [[maybe_unused]] ContainerView {
    protected:
        ContainerType &mContainer;      ///< A reference to the standard container.

    public:
        ContainerView() = delete;

        /**
         * @brief Constructor.
         * @param container The standard container to provide the view of.
         * @param reverse If true reverse the container view, if false do not reverse.
         */
        explicit ContainerView(ContainerType &container) : mContainer(container) {}

        /**
         * @brief Get the begin iterator for the reverse view
         * @return the standard container rbegin iterator
         */
        auto begin() {
            return std::begin(mContainer);
        }

        /**
         * @brief Get the end iterator for the reverse view
         * @return the standard container rend iterator
         */
        auto end() {
            return std::end(mContainer);
        }
    };

    /**
    * @brief Composite a pack of arguments that are streamable to a string.
    * @tparam Arg First argument type
    * @tparam Args Rest of the argument types
    * @param arg First argument
    * @param args Rest of the arguments
    * @return The composited string
    */
    template<typename Arg, typename... Args>
    std::string StringCompositor(Arg &&arg, Args &&... args) {
        std::stringstream out;
        out << std::forward<Arg>(arg);
        ((out << std::forward<Args>(args)), ...);
        return out.str();
    }
}
