/**
 * @file WebCache.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-12-05
 */

#pragma once

#include <algorithm>
#include <cstdint>
#include <filesystem>
#include <future>
#include <iostream>
#include <map>
#include <mutex>
#include <string_view>
#include <utility>
#include <vector>
#include "SDL++.h"
#include "Rose.h"
#include "TimerTick.h"
#include "XDGFilePaths.h"

namespace rose {
    namespace fs = ::std::filesystem;

    using WebCacheProtocol = Protocol<unique_id_type,long>;     ///< WebCache result protocol

    class WebCache;
    class WebRequest;

    /**
     * @class WebRequest
     * @brief An abstraction of a web request to fetch a file from a WebServer instance.
     */
    class WebRequest {
        friend class WebCache;

    protected:
        unique_id_type mWebRequestId{};       ///< The sequence number of this request.
        std::error_code mEc;                  ///< The last error code from file system actions.
        std::string mRequestUrl;              ///< The remainder of the request URL, appended to server base URL.
        fs::path mItemPath;                   ///< The path to store the fetched file, relative to server mCacheDirectory.
        fs::path mTempPath;                   ///< The temporary path used while file is received, generated from mItemPath.
        long responseCode{599};               ///< The response code returned by the server.


        /// The duration local files are considered valid. The interval between cache refresh checks.
        std::chrono::system_clock::duration mCacheValidDuration{};

    public:
        WebRequest() = delete;

        ~WebRequest() = default;

        /**
         * @brief Construct a web request associated with a server.
         * @tparam S A type convertable to std::string.
         * @param webServer The associated web server instance.
         * @param url The local part of the URL.
         * @param path The local part of the cache path.
         */
        template<class S1, class S2>
        WebRequest(S1 url, S2 path, std::chrono::system_clock::duration cacheValidDuration, unique_id_type uniqueId = 0) {
            static_assert(std::is_constructible_v<std::string, S1> && std::is_constructible_v<std::string, S2>,
                          "The url must be convertable to std::string.");
            if (uniqueId == 0)
                uniqueId = UniqueId::userId();
            mWebRequestId = uniqueId;
            mRequestUrl = url;
            mItemPath = path;
            mTempPath = mItemPath;
            auto tempName = mTempPath.filename().string();
            tempName = std::string{"."}.append(tempName);
            mTempPath.replace_filename(fs::path(tempName));
            mCacheValidDuration = cacheValidDuration;
        }

        /**
         * @brief Convert a filesystem time to a system clock time point.
         * @tparam T The type of the file time point.
         * @param fileTimePoint The value of the file time point.
         * @return std::chrono::system_clock::time_point.
         */
        template<typename T>
        static std::chrono::system_clock::time_point fileClockToSystemClock(T fileTimePoint) {
            using namespace std::chrono_literals;
            using namespace std::chrono;

            auto sysWriteTime = time_point_cast<system_clock::duration>(
                    fileTimePoint - decltype(fileTimePoint)::clock::now() +
                    system_clock::now());
            return sysWriteTime;
        }

        /**
         * @brief Get the time the item was cached if the cache time has cacheLoadRequired.
         * @param path The item path (returned by itemLocalPath()).
         * @return A time_t expressing the last write time of the local file if cache time has cacheLoadRequired, an
         * empty std::optional otherwise.
         */
        std::optional<time_t> cacheTime(const fs::path &path) {
            if (fs::exists(path)) {
                auto cacheFiletime = fileClockToSystemClock(std::filesystem::last_write_time(path));
                auto cacheFileAge = std::chrono::system_clock::now() - cacheFiletime;
                if (cacheFileAge.count() > mCacheValidDuration.count())
                    return std::chrono::system_clock::to_time_t(cacheFiletime);
            }
            return std::nullopt;
        }

        /**
         * @brief Test to see if the cached file should be reloaded.
         * @return true if reload required, false otherwise.
         */
        bool cacheLoadRequired() {
            if (fs::exists(mItemPath)) {
                auto cacheFileTime = cacheTime(mItemPath);
                return cacheFileTime.has_value();
            }
            return true;
        }

        std::tuple<unique_id_type,long> fetch();
    };

    /**
     * @class WebCache
     * @brief The web cache manager singleton.
     */
    class WebCache {
    protected:
        /**
         * Which minute out of FetchOnMinuteRange periodic re-fetches occur range [0..FetchOnMinuteRange], negative for
         * fetching on each minute. This is incremented each time the fetch method fails to acquire the lock so that
         * fetching is tried again on the next minute.
         */
#ifdef DEBUG_BUILD
        int mFetchOnMinute{-1};
#endif
#ifdef RELEASE_BUILD
        int mFetchOnMinute{0};
#endif

        static constexpr int FetchOnMinuteRange = 5;

        std::mutex mWebCacheLock{};                 ///< Mutex lock to prevent contention on data.

        std::vector<WebRequest> mRequestList{};     ///< A list of requests.

        /// A list of futures generated by fetching updates and checked once per second.
        std::vector<std::future<std::tuple<unique_id_type ,long>>> mFutures;

        TickProtocol::slot_type minute_slot{};      ///< The Tick slot to trigger periodic fetches.
        TickProtocol::slot_type seconds_slot{};     ///< The Tick slot to reap futures.

        /**
         * @brief Construct the WebCache instance.
         */
        WebCache() {
            // Build the minute Tick receiver.
            minute_slot = TickProtocol::createSlot();
            minute_slot->receiver = [&](int minutes) {
                if (mFetchOnMinute < 0 || minutes % FetchOnMinuteRange == mFetchOnMinute) {
                    if (mWebCacheLock.try_lock()) {
                        if (mFutures.empty()) {
                            fetchAll();
                        } else {
                            // Still processing futures, try again in a minute.
                            mFetchOnMinute = (mFetchOnMinute + 1) % FetchOnMinuteRange;
                        }
                        mWebCacheLock.unlock();
                    } else {
                        // Could not get the lock, try again in a minute.
                        mFetchOnMinute = (mFetchOnMinute + 1) % FetchOnMinuteRange;
                    }
                }
            };

            // Build the second Tick receiver.
            seconds_slot = TickProtocol::createSlot();
            seconds_slot->receiver = [&](int seconds) {
                if (seconds == 0)
                    return;
                if (mWebCacheLock.try_lock()) {
                    if (!mFutures.empty())
                        checkFutures();
                    mWebCacheLock.unlock();
                }
            };

            // Connect the signals.
            TimerTick::get().minuteSignal.connect(minute_slot);
            TimerTick::get().secondSignal.connect(seconds_slot);
        }

    public:
        WebCacheProtocol::signal_type cacheResultSignal{};    ///< The cache result signal.

        void wait() {
            while (!mFutures.empty()) {
                checkFutures();
                std::this_thread::sleep_for(std::chrono::milliseconds(1));
            }
        }

        ~WebCache() = default;

        /**
         * @brief Construct the singleton WebCache and return a reference to it.
         * @return WebCache reference.
         */
        static WebCache& get() {
            static WebCache webCache{};
            return webCache;
        }

        /**
         * @brief Emplace a new web request instance.
         * @tparam S A type that is convertable to std::string.
         * @param serverId The identifier of the associated server instance, returned by WebCache::emplaceServer()
         * @param localUrl The local part of the file URL, appended to siteUrl.
         * @param localPath The local part of the cache file location, appended to cacheDir.
         * @return The identifier of the request instance.
         */
        template<class S1, class S2>
        unique_id_type emplaceRequest(S1 url, S2 path, std::chrono::system_clock::duration cacheValidDuration, unique_id_type uniqueId = 0) {
            const std::lock_guard<std::mutex> lockGuard{mWebCacheLock};
            static_assert(std::is_constructible_v<std::string, S1> && std::is_constructible_v<std::string, S2>,
                          "The url and path must be convertable to std::string.");
            if (uniqueId == 0)
                uniqueId = UniqueId::userId();
            mRequestList.template emplace_back(std::string{url}, std::string{path}, cacheValidDuration, uniqueId);
            auto webRequest = mRequestList.back();
            fs::create_directories(webRequest.mItemPath.parent_path());
            return mRequestList.back().mWebRequestId;
        }

        /**
         * @brief Determine if any of the request files have cacheLoadRequired.
         * @return true if at least one has cacheLoadRequired, false otherwise.
         */
        bool initialFetchRequired() {
            for (auto &request : mRequestList) {
                if (request.cacheLoadRequired())
                    return true;
            }
            return false;
        }

        /**
         * @brief A convenience function to initialize the WebCache data as required on application startup.
         * @details This should be called after the application is built and all components which require
         * cached files from the web have registered their servers and requests but before calling
         * Application::eventLoop()
         */
        void initializeCache() {
            if (initialFetchRequired())
                fetchAll();
        }

        /**
         * @brief Fetch all files except those that are in the cache and have not cacheLoadRequired.
         * @details Files that are cached and not cacheLoadRequired are reported as 304 Not Modified.
         * <p/> The results are reported when the request std::future is checked.
         * <p/> This method should not normally be called by the user program, except at startup
         * to perform an initial fetch of the files. WebCache::initialFetchRequired() can be used
         * to determine if an initial fetch is required.
         */
        void fetchAll() {
            if (mWebCacheLock.try_lock()) {
                for (auto &request: mRequestList) {
                    auto valid = request.cacheLoadRequired();
                    if (valid)
                        mFutures.push_back(std::async(&WebRequest::fetch, &request));
                }
            }
        }

        /**
         * @brief Get the cache file path for a request.
         * @param requestId The identity sequence of the request.
         * @return A std::optional<std::filesystem::path> of the found request path, or empty.
         */
        std::optional<fs::path> requestFilePath(unique_id_type requestId) {
            auto webRequest = std::find_if(mRequestList.begin(), mRequestList.end(), [& requestId](const WebRequest &wr){
                return wr.mWebRequestId == requestId;
            });
            if (webRequest != mRequestList.end())
                return webRequest->mItemPath;
            return std::nullopt;
        }

        template<class S>
        std::optional<long> findRequest(S localUrl) {
            static_assert(std::is_constructible_v<std::string, S>,
                          "Parameter localUrl mst be convertible to std::string.");
            auto requestUrl = std::string(localUrl);
            auto webRequest = std::find_if(mRequestList.begin(), mRequestList.end(),
                                           [&requestUrl](const WebRequest &wr) {
                                               return wr.mRequestUrl == requestUrl;
                                           });
            if (webRequest != mRequestList.end())
                return webRequest->mWebRequestId;
            return std::nullopt;
        }

    protected:
        /**
         * @brief Check each request std::future.
         * @details Completed futures are reaped and the results sent out on a signal. This is checked once
         * pers second automatically.
         */
        void checkFutures() {
            for (auto &f: mFutures) {
                if (f.valid()) {
                    f.wait();
                    auto[requestId, result] = f.get();
                    if (cacheResultSignal)
                        cacheResultSignal.transmit(requestId, result);
                    std::cout << requestId << ',' << result << '\n';
                }
            }
            mFutures.erase(std::remove_if(mFutures.begin(), mFutures.end(),
                                          [](std::future<std::tuple<unique_id_type , long>> &it) -> bool {
                                              return !it.valid();
                                          }), mFutures.end());
        }
    };

    struct WebRequestPreload{
        std::string_view RequestUrl;
        std::string_view LocalPath;
    };

    template<size_t N>
    struct WebCachePreload {
        std::string_view ServerUrl;
        std::string_view RelativePath;
        std::chrono::system_clock::duration cacheValidDuration{};
        std::array<WebRequestPreload,N> Requests;
    };

    struct WebCachePreloadResult {
        bool waitForCache{false};
        std::vector<std::vector<unique_id_type>> list{};

        void addResult(std::tuple<bool, std::vector<unique_id_type>> result) {
            waitForCache |= std::get<0>(result);
            list.push_back(std::get<1>(result));
        }
    };

    template<size_t N>
    std::tuple<bool,std::vector<unique_id_type>> webCachePreloadItem(WebCachePreload<N>& arg) {
        bool waitForCache = false;
        SingletonGuard<XDGFilePaths> filePaths{};
        SingletonGuard<WebCache> webCache{};

        auto [found,path] = filePaths->findFilePath(XDGFilePaths::XDG_CACHE_HOME, arg.RelativePath);
        waitForCache |= !found;
        std::vector<unique_id_type> requestIdList{};
        for (auto& request : arg.Requests) {
            auto fullPath = path.string() + std::string(request.LocalPath);
            auto fullUrl = std::string(arg.ServerUrl) + '/' + std::string(request.RequestUrl);
            requestIdList.push_back(webCache->emplaceRequest(fullUrl, fullPath, arg.cacheValidDuration));
            auto localPath = path;
            localPath.append(std::string(request.LocalPath));
            waitForCache |= !fs::exists(localPath);
        }
        return std::make_tuple(waitForCache,requestIdList);
    }

    template<class...Args>
    WebCachePreloadResult webCachePreload(Args...args) {
        WebCachePreloadResult result{};
        std::vector<std::vector<unique_id_type>> requestLists{};
        (result.addResult(webCachePreloadItem(args)),...);
        return result;
    }
}

