/**
 * @file Widget.h
 * @author Richard Buckley <richard.buckley@ieee.org>
 * @version 1.0
 * @date 2021-11-22
 */

#pragma once

#include <atomic>
#include <filesystem>
#include <functional>
#include <memory>
#include <utility>
#include <vector>
#include <map>
#include <exception>
#include <iostream>
#include "Color.h"
#include "Event.h"
#include "Sdl++/Metrics.h"
#include "Property.h"
#include "Theme.h"

namespace rose {

    class Widget;
    namespace gm {
        class Context;
    }

    enum class LayoutHintType : std::size_t {
        HORIZONTAL_ALIGNMENT,
        VERTICAL_ALIGNMENT,
        IS_ELASTIC,
    };

    enum class HorizontalAlignment : std::size_t {
        NONE, CENTER, LEFT, RIGHT,
    };

    enum class VerticalAlignment : std::size_t {
        NONE, CENTER, TOP, BOTTOM,
    };

    enum class ElasticAxis : std::size_t {
        NONE, HORIZONTAL, VERTICAL, BOTH, ASPECT_RATIO,
    };

    using LayoutHintValue = std::variant<bool, int, float, HorizontalAlignment, VerticalAlignment, ElasticAxis>;

    class LayoutHint : public std::pair<LayoutHintType, LayoutHintValue> {
    public:
        using base_class = std::pair<LayoutHintType, LayoutHintValue>;

        LayoutHint() = default;

        ~LayoutHint() = default;

        template<class ValueType>
        LayoutHint(LayoutHintType type, ValueType value) : base_class(type,value) {
            switch(type) {
                case LayoutHintType::HORIZONTAL_ALIGNMENT:
                    if (std::holds_alternative<HorizontalAlignment>(second))
                        return;
                    break;
                case LayoutHintType::VERTICAL_ALIGNMENT:
                    if (std::holds_alternative<VerticalAlignment>(second))
                        return;
                    break;
                case LayoutHintType::IS_ELASTIC:
                    if (std::holds_alternative<ElasticAxis>(second))
                        return;
                    break;
            }
            throw std::invalid_argument("LayoutHint type does not match value.");
        }
    };

    /**
     * @class WidgetTreeException
     * @brief An exception thrown on Widget tree logic errors.
     */
    class WidgetTreeException : public std::logic_error {
    public:
        [[maybe_unused]] explicit WidgetTreeException(const char *what) : std::logic_error(what) {}

        [[maybe_unused]] explicit WidgetTreeException(const std::string &what) : std::logic_error(what) {}
    };

    /**
     * @class Gadget
     * @brief The simplest member of the Gadget tree. Provides no functionality other than
     * the ability to be a member of the tree.
     */
    class Gadget : public std::enable_shared_from_this<Gadget> {
    protected:
        std::shared_ptr<Gadget> mTrunk{};    ///< The node immediately above this in the tree.

    public:
        Gadget() = default;

        virtual ~Gadget() = default;

        /**
         * @brief Find the root of the Gadget tree from this.
         * @return A std::shared_ptr to the root of the tree, or a if this Gadget is not connected to the trunk.
         */
        [[maybe_unused]] [[nodiscard]] std::shared_ptr<Gadget> root() {
            auto node = shared_from_this();
            while (node->mTrunk) {
                node = node->mTrunk;
            }
            return node;
        }

        auto trunk() const {
            return mTrunk;
        }

        /**
         * @brief Add the trunk for this Gadget.
         * @param trunkGadget The trunk. Throws GadgetTreeException if the trunk is already set.
         */
        void setTrunk(std::shared_ptr<Gadget> trunkGadget) {
            if (mTrunk)
                throw WidgetTreeException("Trunk already set for Gadget.");
            mTrunk = std::move(trunkGadget);
        }

        /**
         * @brief Unset the trunk for this Gadget.
         */
        [[maybe_unused]] void unsetTrunk() {
            mTrunk = nullptr;
        }
    };

    /**
     * @class Bag
     * @brief A container to hold a set of Gadgets to add functionality to an object.
     * @tparam ContentType the type of content the Bag can hold, must be derived from Gadget, defaults to Gadget.
     */
    template<typename ContentType = Gadget>
    class [[maybe_unused]] Bag : public Gadget {
    public:
        static_assert(std::is_base_of_v<Gadget,ContentType>, "Bag content must derive from Gadget.");
        using content_type = ContentType;
        using container_type = std::vector<std::shared_ptr<content_type>>;
        using size_type [[maybe_unused]] = typename container_type::size_type;

    protected:
        container_type mContents{};

    public:

        Bag() = default;

        ~Bag() override {
            for (const auto& item : mContents) {
                item->unsetTrunk();
            }
            mContents.clear();
            unsetTrunk();
        }

        /**
         * @brief Add an item to the Bag. The item trunk is set to host so gadgets can find their host if necessary.
         * @param host Pointer to the Gadget that is hosting the Bag.
         * @param item Item to add to the Bag.
         */
        [[maybe_unused]] virtual void addItem(const std::shared_ptr<content_type>& item) {
            item->setTrunk(mTrunk);
            mContents.push_back(item);
        }

        /**
         * @brief Access the bag container.
         * @return A reference to the container.
         */
        container_type &view() { return mContents; }
    };

    struct Border {
        int border;
    };

    using WidgetLayoutData = std::tuple<Rectangle<int>, Rectangle<int>, Rectangle<int>, Rectangle<int>, Rectangle<int>>;

    class Screen;
    class Popup;

    /**
     * @struct KeyShortcut
     * @brief Information pertaining to a key event which may be used as an application short cut.
     * @details Each key event is packaged up as a KeyShortcut. The Application shortcut list is
     * searched for a match (defined as keycode values equal and mod share at least one bit).
     */
    struct KeyShortcut {
        bool state{false};                          ///< The key state true for pressed false for released.
        bool repeat{false};                         ///< True if the key is a repeat.
        SDL_Keycode keycode{};                      ///< The virtual key code see SDL_Keycode.
        uint16_t mod{};                             ///< The key modifiers.
    };

    using KeyShortcutFunction = std::function<bool(const KeyShortcut &)>;
    using KeyShortcutPair = std::pair<KeyShortcut, KeyShortcutFunction>;

    class KeyShortcutList : public std::vector<KeyShortcutPair> {
    public:
        KeyShortcutList() = default;

        ~KeyShortcutList() = default;

        std::optional<KeyShortcutPair> findKeyShortcut(const KeyShortcut &keyShortcut);

        bool setKeyShortcut(const KeyShortcut &keyShortcut, const KeyShortcutFunction &function);

    };

    /**
     * @class Widget
     * @brief The base class of all functional UI objects.
     */
    class Widget : public std::enable_shared_from_this<Widget> {
    public:
        static constexpr std::string_view WidgetClass{"Widget"};
        static constexpr std::string_view ActiveName = "Active";
        static constexpr std::string_view VisibleName = "Visible";
        static constexpr std::string_view SensitiveName = "Sensitive";
        static constexpr std::string_view AbortOnLeaveName = "AbortOnLeave";
        static constexpr std::string_view IconOnLeftName = "IconOnLeft";
        static constexpr std::string_view BorderName = "Border";
        static constexpr std::string_view MarginName = "Margin";
        static constexpr std::string_view PaddingName = "Padding";
        static constexpr std::string_view BaseColorName = "BaseColor";
        static constexpr std::string_view TopColorName = "TopColor";
        static constexpr std::string_view BotColorName = "BotColor";
        static constexpr std::string_view LeftColorName = "LeftColor";
        static constexpr std::string_view RightColorName = "RightColor";
        static constexpr std::string_view InvertColorName = "InvertColor";
        static constexpr std::string_view TextColorName = "TextColor";

        enum WidgetFlags {
            ACTIVE_ON,
            ACTIVE_OFF,
            SENSITIVE_ON,
            SENSITIVE_OFF,
            VISIBLE_ON,
            VISIBLE_OFF,
            MODAL_ON,
            MODAL_OFF,
        };

    private:
        /**
         * @brief Get any resources from the resource databases for the Widget and the ultimate Derived Widget.
         * @details The order of the databases in the ResourceSet and the order of the search names processed
         * set which values take precedence.
         */
        void getResources();

    protected:
        Theme &mTheme;
        std::shared_ptr<Widget> mTrunk{};          ///< The node immediately above this in the tree.
        std::map<property_key, property_value> mProperties{};
        std::vector<LayoutHint> mLayoutHints{};
        Padding<int> mPadding{0};              ///< Padding around the content.
        int mBorder{0};                           ///< The size of the border, or zero for no border.

        /**
         * The following values: mPaddingRect through mContentRect are in absolute screen coordinates.
         */
        Rectangle<int> mPaddingRect{},            ///< The rectangle containing the padded widget.
                mBorderRect{},                    ///< The rectangle containing the bordered widget.
                mMarginRect{},                    ///< The rectangle containing the margined widget.
                mContentRect{};                   ///< The rectangle containing the content of the widget.
        Margin<int> mMargin{0};                ///< Space between the border and the content.
        bool mVisible{true};                      ///< True when the Widget will be rendered.
        bool mActive{false};                      ///< True when the Widget is ON.
        bool mSensitive{true};                    ///< True when the Widget will respond to user interaction.
        bool mModal{false};                       ///< True when widget should be modal, only useful for popups widget.
        bool mAbortOnLeave{false};                ///< How to handle two step interactions when the cursor leaves the Widget.
        bool mIconOnLeft{true};                   ///< When Icons and Text are used together, put the Icon on the left.
        bool mPerFrameProcess{false};             ///< When true this Widget::perFrameProcess() will be called.

        KeyShortcutList mKeyShortcutList{};

        SDL_Color mBaseColor{}, mTopColor{}, mBotColor{}, mLeftColor{}, mRightColor{}, mInvertColor{}, mTextColor{};

    public:
        Widget() : mTheme(Theme::get()) {
            property_key key = property::Class;
            property_value value = WidgetClass;
            mProperties.emplace(key, value);
            getResources();
        };

        Widget(const property_key key, const property_value &value) : mTheme(Theme::get()) {
            mProperties.emplace(key, value);
            getResources();
        }

        virtual ~Widget() = default;

        /**
         * @brief Called by the builder after all the parameter arguments have been processed.
         */
        virtual void buildComplete();

        /**
         * @brief Called by the builder after the widget has been inserted in the tree.
         */
        virtual void insertedInTree();

        virtual void debugParameters();

        /**
         * @brief Handle events.
         * @param event The event
         * @return true if the widget used the event, false if it did not.
         */
        virtual bool handleEvent(const Event& event);

        /**
         * Deprecated.
         */
        virtual bool takeFocus(const Event &e);

        /**
         * @brief Determine if this Widget currently needs processing at each frame.
         * @details If true the function perFrameProcess will be called.
         * @return true if per frame processing is needed.
         */
        bool isPerFrameProcess() const { return mPerFrameProcess; }

        /**
         * @brief Set the per frame processing flag to true.
         */
        virtual void setPerFrameProcess();

        /**
         * @brief Set the per frame processing flag to false.
         */
        virtual void clearPerFrameProcess();

        /**
         * @brief This method containsRelative the code the Widget must execut per fram.
         */
        virtual void perFrameProcess();

        /**
         * @brief Determine if the Widget containsRelative a position.
         * @details The position is relative to the origin of any padding the widget has, but positions
         * within the padding area are not considered inside the Widget.
         * @param pos The relative position.
         * @return true if the position is inside the widget.
         */
        bool containsRelative(const Position<int>& pos) {
            Rectangle<int> rect{position + mPadding.position(), size + mMargin.size() + Size(mBorder*2)};
            return rect.contains(pos);
        }

        /**
         * @brief Determines if the Widget containsRelative a position.
         * @details The position is an absolute screen position. A position within the Widget padding is not
         * considered inside the widget. This method depends on values calculated during layout and again
         * before drawing so they can not be relied upon to be valid at all time.
         * @param pos The absolute position.
         * @return true if the position is inside the widget.
         */
        bool containsAbsolute(const Position<int>& pos) {
            return mBorderRect.contains(pos);
        }

        auto getLayoutHints() const { return mLayoutHints; }

        ElasticAxis getLayoutHintElasticAxis() const;

        SDL_Color getBaseColor() const {
            return mBaseColor;
        }

        /**
         * @brief Get a pointer to a WidgetClass from a Widget.
         * @tparam WidgetClass The class to get a pointer to.
         * @return A std::shared_ptr<WidgetClass> if WidgetClass is the derived class of this Widget, nullptr otherwise.
         */
        template<class WidgetClass>
        std::shared_ptr<WidgetClass> as() {
            static_assert(std::is_base_of_v<Widget,WidgetClass>, "WidgetClass must be derived from Widget.");
            return std::dynamic_pointer_cast<WidgetClass>(shared_from_this());
        }

        [[nodiscard]] constexpr auto getMargin() const noexcept { return mMargin; }

        [[nodiscard]] constexpr bool isModal() const noexcept { return mModal; }

        /**
         * @brief Type directed parameter setting.
         * @details Derived classes must implement this function to take parameter types that they handle, then
         * call this function on their base class for those parameter types they don't handle. Any parameter
         * types that reach this level and are not handled by Widget will trigger a static assertion.
         * @tparam Parameter The type of the parameter.
         * @param parameter The value of the parameter.
         */
        template<class Parameter>
        void setParameter(Parameter) {
            throw std::runtime_error("Parameter not handled by Widget");
        }

        /**
         * @brief Find the root of the Widget tree from this.
         * @return A std::shared_ptr to the root of the tree, or a if this Widget is not connected to the trunk.
         */
        [[maybe_unused]] [[nodiscard]] std::shared_ptr<Widget> root() {
            auto node = shared_from_this();
            while (node->mTrunk) {
                node = node->mTrunk;
            }
            return node;
        }

        [[nodiscard]] auto trunk() const noexcept {
            return mTrunk;
        }

        /**
         * @brief Add the trunk for this Widget.
         * @param trunkWidget The trunk. Throws WidgetTreeException if the trunk is already set.
         */
        void setTrunk(std::shared_ptr<Widget> trunkWidget) {
            if (mTrunk)
                throw WidgetTreeException("Trunk already set for Widget.");
            mTrunk = std::move(trunkWidget);
        }

        /**
         * @brief Unset the trunk for this Widget.
         */
        [[maybe_unused]] void unsetTrunk() {
            mTrunk = nullptr;
        }

        [[nodiscard]] std::optional<property_value> findLocalProperty(property_key key) const;

        [[nodiscard]] std::optional<property_value> findProperty(property_key, bool searchTree = false);

        /**
         * @brief Set a property on the Widget.
         * @param property The property which consists of a key, value pair.
         */
        void setProperty(const Property& property) {
            mProperties[property.first] = property.second;
        }

        /**
         * @brief Set the visibility of the Widget.
         * @details Widgets which are not visible are not laid out and are not drawn.
         * @param visible The new visibility setting.
         */
        void setVisible(bool visible) noexcept;

        Size<int> size{};               ///< The size of the widget content.
        Position<int> position{};       ///< The position of the widget relative to its container.

        /**
         * @brief Draw the widget. The exact implementation depends on the specific widget.
         * @param context The graphics context used to draw the widget elements.
         * @param theme The theme used to select colors, fonts, etc.
         * @param trunkPosition The Position of the trunk container relative to its trunk.
         * @param trunkSize The size of the trunk widget, if known.
         */
        virtual void
        draw(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize);

        /**
         * @brief Lay out the widget. The exact implementation depends on the specific widget.
         * @param context The graphics context used to draw the widget elements.
         * @param theme The theme used to select colors, fonts, etc.
         * @param trunkPosition The Position of the trunk container relative to its trunk.
         * @param trunkSize The size of the trunk widget, if known.
         */
        virtual void
        layout(gm::Context &context, const Position<int> &trunkPosition, const Size<int> &trunkSize);

        /**
         * @brief Get a pointer to the screen.
         * @return std::shared_ptr<Screen>
         */
        [[nodiscard]] std::shared_ptr<Screen> screen() const;

        /**
         * @brief Indicate that this widget needs to be redrawn during the next frame.
         */
        void redrawBackground() const;

        /**
         * @brief Compute the widget layout data.
         * @details Layout data include the screen absolute rectangles used to draw the widget. Of the five
         * rectangles returned all except the trunk rectangle are guaranteed to be nested and in the order
         * of largest to smallest.
         * @param trunkPosition The position of the widget trunk relative to its trunk.
         * @param trunkSize The size of the widget trunk, if known.
         * @return A std::tuple with five rectangles: the trunk, padding, boarder, margin, and content.
         */
        WidgetLayoutData widgetLayoutData(const Position<int>& trunkPosition, const Size<int> &trunkSize);

        /**
         * @brief Get the layout size of the widget.
         * @details The layout size is the total screen size of the widget including margin, border and padding.
         * @return Size<int>
         */
        [[nodiscard]] Size<int> getLayoutSize() const noexcept;

        /**
         * @brief Set the layout size of the widget.
         * @details The layout size is the total screen size of the widget including margin, border and padding.
         * @param layoutSize The new layout size.
         */
        void setLayoutSize(const Size<int> &layoutSize);

        /**
         * @brief Draw the widget background and border, return the background color.
         * @details The widget background is filled, then if the border size is non-zero a border is drawn according
         * to the theme and widget state (sensitive & active).
         * @param context The graphics context used to draw.
         * @return The widget background color.
         */
        SDL_Color drawBorderBackground(gm::Context &context) const;

        /**
         * @brief Search the widget local list of shortcut key specifications and invoke associated code.
         * @param keyShortcut The key shortcut specification.
         * @return true if the widget implements the shortcut.
         */
        virtual bool handleKeyShortcut(const KeyShortcut &keyShortcut) {
            if (!mKeyShortcutList.empty())
                if (auto keyShortcutPair = mKeyShortcutList.findKeyShortcut(keyShortcut); keyShortcutPair) {
                    return keyShortcutPair->second(keyShortcut);
                }
            return false;
        }
    };

    template<>
    inline void Widget::setParameter<Widget::WidgetFlags>(Widget::WidgetFlags parameter) {
        switch (parameter) {
            case ACTIVE_ON:
                mActive = true;
                break;
            case ACTIVE_OFF:
                mActive = false;
                break;
            case SENSITIVE_ON:
                mSensitive = true;
                break;
            case SENSITIVE_OFF:
                mSensitive = false;
                break;
            case VISIBLE_ON:
                mVisible = true;
                break;
            case VISIBLE_OFF:
                mVisible = false;
                break;
            case MODAL_OFF:
                mModal = false;
                break;
            case MODAL_ON:
                mModal = true;
                break;
        }
    }

    template<>
    inline void Widget::setParameter<LayoutHint>(LayoutHint parameter) {
        mLayoutHints.push_back(std::move(parameter));
    }

    template<>
    inline void Widget::setParameter<Position<int>>(Position<int> parameter) {
        position = std::move(parameter);
    }

    template<>
    inline void Widget::setParameter<Size<int>>(Size<int> parameter) {
        size = std::move(parameter);
    }

    template <>
    inline void Widget::setParameter<Property>(Property parameter) {
        mProperties.emplace(std::move(parameter));
    }

    template <>
    inline void Widget::setParameter<Padding<int>>(Padding<int> parameter) {
        mPadding = parameter;
    }

    template <>
    inline void Widget::setParameter<Margin<int>>(Margin<int> parameter) {
        mMargin = parameter;
    }

    template <>
    inline void Widget::setParameter<Border>(Border parameter) {
        mBorder = parameter.border;
    }

    using WidgetPtr = std::shared_ptr<Widget>;

}

/**
 * @brief Add a property value to a Widget pointer
 * @tparam WidgetType The type of Widget pointer
 * @param widget The Widget pointer
 * @param property The Property to add.
 * @return The Widget pointer.
 */
template<typename WidgetType>
inline auto operator + (const std::shared_ptr<WidgetType>& widget, const rose::Property& property) {
    static_assert(std::is_base_of_v<rose::Widget,WidgetType>, "Properties can only be added to rose::Widget objects.");
    widget->setProperty(property);
    return widget;
}
